# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Correlations in the optimization database
*****************************************
"""

from __future__ import absolute_import, division, unicode_literals

from os.path import basename, dirname, join, splitext

import matplotlib.gridspec as gridspec
import numpy as np
import pylab
from future import standard_library
from matplotlib import ticker
from numpy import atleast_2d

from gemseo.post.opt_post_processor import OptPostProcessor

standard_library.install_aliases()


from gemseo import LOGGER


class Correlations(OptPostProcessor):
    """
    The **Correlations** post processing
    builds scatter plots of correlated variables among design
    variables, outputs functions and constraints

    The plot method considers all variable correlations
    greater than 95%. An other level value, a sublist of variable names
    or both can be passed as options. The x- and y- figure sizes
    can also be modified in option.
    It is possible either to save the plot, to show the plot or both.
    """

    def _run(self, **options):
        """Visualizes the optimization history

        :param options: options for the post processing,
            see associated JSON file
        """
        functions = self.opt_problem.get_all_functions()
        func_names = [func.name for func in functions]
        self._plot(func_names, **options)

    def _plot(
        self,
        func_names=None,
        coeff_limit=0.95,
        n_plots_x=5,
        n_plots_y=5,
        save=False,
        show=False,
        file_path=None,
        extension="pdf",
    ):
        """
        Plots the correlations graph

        :param coeff_limit: if the correlation between the variables
            is lower than coeff_limit, the plot is not made
        :type coeff_limit: bool
        :param show: if True, displays the plot windows
        :type show: bool
        :param save: if True, exports plot to pdf
        :type save: bool
        :param file_path: the base paths of the files to export
        :type file_path: str
        :param func_names: the func_names on which correlations is computed
        :type func_names: list(str)
        :param n_plots_x: number of horizontal plots
        :type n_plots_x: int
        :param n_plots_y: number of vertical plots
        :type n_plots_y: int
        :param extension: file extension
        :type extension: str
        """
        n_slide = 0
        values_array, variables_names, _ = self.database.get_history_array(
            func_names, None, True, 0.0
        )
        corr_coeffs_array = self.__compute_correlations(values_array)
        i_corr, j_corr = np.where(
            (np.abs(corr_coeffs_array) > coeff_limit)
            & (np.abs(corr_coeffs_array) < (1.0 - 1e-9))
        )
        LOGGER.info("Detected %s correlations > %s", i_corr.size, coeff_limit)
        if i_corr.size <= 16:
            n_plots_x = 4
            n_plots_y = 4
        figs = []
        spec = gridspec.GridSpec(n_plots_y, n_plots_x, wspace=0.3, hspace=0.75)
        spec.update(top=0.95, bottom=0.06, left=0.08, right=0.95)
        fig = None
        fig_indx = 0
        if file_path is not None:
            root = splitext(file_path)[0]
            root_dir = dirname(root)
            base_n = basename(root)
        else:
            root_dir = "."
            base_n = ""
        for plot_index, (i, j) in enumerate(zip(i_corr, j_corr)):
            plot_index_loc = plot_index % (n_plots_x * n_plots_y)
            if plot_index_loc == 0:
                if fig is not None:  # Save previous plot
                    fig_indx += 1
                    base_loc = base_n + "correlations_" + str(fig_indx)
                    fpath = join(root_dir, base_loc)
                    self._save_and_show(
                        fig, file_path=fpath, save=save, show=show, extension=extension
                    )
                    pylab.plt.close(fig)
                fig = pylab.plt.figure()
                figs.append(fig)
                mng = pylab.plt.get_current_fig_manager()
                mng.resize(1200, 900)
                ticker.MaxNLocator(nbins=3)

            # plt.suptitle('All variables are normalized')
            self.__create_sub_correlation_plot(
                i,
                j,
                corr_coeffs_array[i, j],
                fig,
                spec,
                plot_index_loc,
                n_plots_y,
                n_plots_x,
                values_array,
                variables_names,
            )
        if fig is not None:
            base_loc = base_n + "correlations_" + str(fig_indx + 1)
            fpath = join(root_dir, base_loc)
            self._save_and_show(
                fig, save=save, show=show, file_path=fpath, extension=extension
            )
            pylab.plt.close(fig)
        return n_slide

    def __create_sub_correlation_plot(
        self,
        i_ind,
        j_ind,
        corr_coeff,
        fig,
        spec,
        plot_index,
        n_plot_v,
        n_plot_h,
        values_array,
        variables_names,
    ):
        """Creates a correlation plot"""
        gs_curr = spec[int(plot_index / n_plot_v), plot_index % n_plot_h]
        ax1 = fig.add_subplot(gs_curr)
        x_plt = values_array[:, i_ind]
        y_plt = values_array[:, j_ind]
        ax1.scatter(x_plt, y_plt, c="b", s=30)
        self.out_data_dict[(i_ind, j_ind)] = (
            variables_names[i_ind],
            variables_names[j_ind],
            corr_coeff,
        )
        ax1.set_xlabel(variables_names[i_ind], fontsize=9)
        # Update y labels spacing
        start, stop = ax1.get_ylim()
        ax1.yaxis.set_ticks(np.arange(start, stop, 0.24999999 * (stop - start)))
        start, stop = ax1.get_xlim()
        ax1.xaxis.set_ticks(np.arange(start, stop, 0.24999999 * (stop - start)))
        ax1.set_ylabel(variables_names[j_ind], fontsize=10)
        ax1.tick_params(labelsize=10)
        ax1.set_title("R=%5f" % corr_coeff, fontsize=12)
        ax1.grid()

    @classmethod
    def __compute_correlations(cls, values_array):
        """Compute correlations"""
        ccoeff = np.corrcoef(values_array.astype(float), rowvar=False)
        return np.tril(atleast_2d(ccoeff))  # Keep upper diagonal only



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Display a Pareto Front
**********************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from future import standard_library

from gemseo.algos.pareto_front import generate_pareto_plots
from gemseo.post.opt_post_processor import OptPostProcessor

standard_library.install_aliases()


from gemseo import LOGGER


class ParetoFront(OptPostProcessor):
    """
    Compute the Pareto front
    Search for all non dominated points, ie there exists j such that
    there is no lower value for obj_values[:,j] that does not degrade
    at least one other objective  obj_values[:,i].

    Generates a plot or a matrix of plots if there are more than 2 objectives.
    Plots in red the locally non dominated points
    for the currrent two objectives.
    Plot in green the globally (all objectives) Pareto optimal points.
    """

    def _plot(
        self,
        objectives=None,
        objectives_labels=None,
        figsize_x=10,
        figsize_y=10,
        show=False,
        save=False,
        file_path=None,
        extension="pdf",
    ):
        """
        Plots the Pareto front

        :param objectives: the functions names or design variables to plot
            if None, use the objective function (may be a vector)
        :type objectives: list(str)
        :param figsize_x: size of figure in horizontal direction (inches)
        :type figsize_x: int
        :param figsize_y: size of figure in vertical direction (inches)
        :type figsize_y: int
        :param show: if True, displays the plot windows
        :type show: bool
        :param save: if True, exports plot to pdf
        :type save: bool
        :param file_path: the base paths of the files to export
        :type file_path: str
        :param extension: file extension
        :type extension: str
        """
        if objectives is None:
            objectives = [self.opt_problem.objective.name]
        add_dv = False
        all_funcs = self.opt_problem.get_all_functions_names()
        all_dv_names = self.opt_problem.design_space.variables_names
        design_variables = None
        if not objectives:
            # function list only contains design variables
            vals = self.database.get_x_history()
            vname = self.database.set_dv_names(vals[0].shape[0])
        else:
            design_variables = []
            for func in list(objectives):
                if func not in all_funcs and func not in all_dv_names:
                    min_f = "-" + func == self.opt_problem.objective.name
                    if min_f and not self.opt_problem.minimize_objective:
                        objectives[objectives.index(func)] = "-" + func
                    else:
                        msg = "Cannot build Pareto front,"
                        msg += " Function " + func + " is neither among"
                        msg += " optimization problem functions :"
                        msg += str(all_funcs) + "nor design variables :"
                        msg += str(all_dv_names)
                        raise ValueError(msg)
                if func in self.opt_problem.design_space.variables_names:
                    # if given function is a design variable, then remove it
                    add_dv = True
                    objectives.remove(func)
                    design_variables.append(func)
            if design_variables == []:
                design_variables = None
            vals, vname, _ = self.database.get_history_array(
                objectives, design_variables, add_dv=add_dv
            )
        if objectives_labels is not None:
            assert len(vname) == len(objectives_labels)
            vname = objectives_labels
        fig = generate_pareto_plots(vals, vname, figsize=(figsize_x, figsize_y))

        self._save_and_show(
            fig, save=save, show=show, file_path=file_path, extension=extension
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Composed distribution based on scipy
====================================

Overview
--------

The :class:`.SPComposedDistribution` class is a concrete class
inheriting from :class:`.ComposedDistribution` which is an abstract one.

Construction
------------

The :class:`.SPComposedDistribution` of a list of given uncertain variables
is built from a list of :class:`.SPDistribution` objects
implementing the probability distributions of these variables
based on the OpenTURNS library and from a copula name.

.. note::

   A copula is a mathematical function used to define the dependence
   between random variables from their cumulative density functions.
   `See more <https://en.wikipedia.org/wiki/Copula_(probability_theory)>`__.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import array

from gemseo.uncertainty.distributions.distribution import ComposedDistribution

standard_library.install_aliases()


class SPComposedDistribution(ComposedDistribution):
    """ Scipy composed distribution. """

    COPULA = {ComposedDistribution.INDEPENDENT_COPULA: None}

    def __init__(self, distributions, copula=ComposedDistribution.INDEPENDENT_COPULA):
        """Constructor.

        :param list(SPDistribution) distributions: list of OT distributions.
        :param str copula: copula name.
        """
        super(SPComposedDistribution, self).__init__(distributions, copula)
        self.distribution = distributions
        self._mapping = {}
        index = 0
        for id_marg, marginal in enumerate(self.distribution):
            for id_submarg in range(marginal.dimension):
                self._mapping[index] = (id_marg, id_submarg)
                index += 1
        self._set_bounds(distributions)

    def cdf(self, vector):
        """Evaluate the cumulative density functions of the marginals
        of a random variable for a given instance.

        :param array vector: instance of the random variable.
        """
        tmp = []
        for index, value in enumerate(vector):
            id1 = self._mapping[index][0]
            id2 = self._mapping[index][1]
            tmp.append(self.distribution[id1].marginals[id2].cdf(value))
        return array(tmp)

    def inverse_cdf(self, vector):
        """Evaluate the inverses of the cumulative density functions of the
        marginals of a random variable for a given vector .

        :param array vector: vector of values comprised between 0 and 1 with
            same dimension as the random variable.
        """
        tmp = []
        for index, value in enumerate(vector):
            id1 = self._mapping[index][0]
            id2 = self._mapping[index][1]
            tmp.append(self.distribution[id1].marginals[id2].ppf(value))
        return array(tmp)

    def _pdf(self, index):
        id1 = self._mapping[index][0]
        id2 = self._mapping[index][1]

        def pdf(level):
            return self.distribution[id1].marginals[id2].pdf(level)

        return pdf

    def _cdf(self, index):
        id1 = self._mapping[index][0]
        id2 = self._mapping[index][1]

        def cdf(level):
            return self.distribution[id1].marginals[id2].cdf(level)

        return cdf



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A formulation for uncoupled or weakly coupled problems
******************************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import super

from future import standard_library

from gemseo.core.chain import MDOChain
from gemseo.core.execution_sequence import ExecutionSequenceFactory
from gemseo.core.formulation import MDOFormulation
from gemseo.utils.data_conversion import DataConversion

standard_library.install_aliases()


class DisciplinaryOpt(MDOFormulation):
    """
    The disciplinary optimization formulation draws the architecture
    of a mono disciplinary
    optimization process from an ordered list of disciplines,
    an objective function and a design space. The objective function
    is minimized by default.
    """

    def __init__(
        self, disciplines, objective_name, design_space, maximize_objective=False
    ):
        """
        Constructor, initializes the objective functions and constraints

        :param disciplines: the disciplines list.
        :type disciplines: list(MDODiscipline)
        :param objective_name: the objective function data name.
        :type objective_name: str
        :param design_space: the design space.
        :type design_space: DesignSpace
        :param maximize_objective: if True, the objective function
            is maximized, by default, a minimization is performed.
        :type maximize_objective: bool
        """
        self.chain = None
        if len(disciplines) > 1:
            self.chain = MDOChain(disciplines)
        super(DisciplinaryOpt, self).__init__(
            disciplines,
            objective_name,
            design_space,
            maximize_objective=maximize_objective,
        )
        self._filter_design_space()
        self._set_defaultinputs_from_ds()
        # Build the objective from its objective name
        self._build_objective_from_disc(objective_name)

    def get_expected_workflow(self):
        """
        Returns the expected execution sequence,
        used for xdsm representation
        """
        if self.chain is None:
            return ExecutionSequenceFactory.serial(self.disciplines[0])
        return self.chain.get_expected_workflow()

    def get_expected_dataflow(self):
        """
        Returns the expected data exchange sequence,
        used for xdsm representation
        """
        if self.chain is None:
            return []
        return self.chain.get_expected_dataflow()

    def get_top_level_disc(self):
        """Returns the disciplines which inputs are required to run the
        associated scenario
        By default, returns all disciplines
        To be overloaded by subclasses

        :returns: the list of top level disciplines
        """
        if self.chain is not None:
            return [self.chain]
        return self.disciplines

    def _filter_design_space(self):
        """
        Filters the design space to keep only available variables
        """
        all_inpts = DataConversion.get_all_inputs(self.get_top_level_disc())
        kept = set(self.design_space.variables_names) & set(all_inpts)
        self.design_space.filter(kept)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#          Arthur Piat: greatly improve the N2 layout
"""
Coupled problem analysis, weak/strong coupling computation using graphs
***********************************************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
from shutil import move

import matplotlib.pyplot as plt
from future import standard_library
from graphviz import Digraph
from numpy import array
from pylab import gca

from gemseo.utils.py23_compat import string_types

standard_library.install_aliases()


from gemseo import LOGGER


class MDOCouplingStructure(object):
    """Structure of the couplings between disciplines
    The methods of this class include the computation of weak,
    strong or all couplings."""

    def __init__(self, disciplines):
        """
        Constructor

        :param disciplines: list of MDO disciplines that possibly
            exchange coupling variables
        """
        self.disciplines = disciplines
        # generate the directed graph of the disciplines and the
        # resulting execution sequence
        self.graph = DependencyGraph(disciplines)
        self.sequence = self._compute_execution_sequence()

    def _compute_execution_sequence(self):
        """Generates the execution sequence of the disciplines.
        Transforms the sequence of indices into a sequence of
        disciplines."""
        sequence = []
        for parallel_tasks in self.graph.execution_sequence:
            parallel_tasks_disc = []
            for component in parallel_tasks:
                # replace index by corresponding discipline
                component_disc = tuple(self.disciplines[index] for index in component)
                parallel_tasks_disc.append(component_disc)
            sequence.append(parallel_tasks_disc)
        return sequence

    @staticmethod
    def is_self_coupled(discipline):
        """
        Tests if the discipline is self coupled
        ie if one of its outputs is also an input

        :param discipline: the discipline
        :returns: a boolean
        """
        return (
            len(
                set(discipline.get_input_data_names())
                & set(discipline.get_output_data_names())
            )
            > 0
        )

    # methods that determine strong/weak/all couplings

    def strongly_coupled_disciplines(self):
        """Determines the strongly coupled disciplines, that is
        the disciplines that occur in (possibly different) MDAs."""
        strong_disciplines = []
        for parallel_tasks in self.sequence:
            for component in parallel_tasks:
                # find MDAs
                if len(component) > 1:
                    for discipline in component:
                        strong_disciplines.append(discipline)
                else:
                    for discipline in component:
                        if self.is_self_coupled(discipline):
                            strong_disciplines.append(discipline)
        return strong_disciplines

    def weakly_coupled_disciplines(self):
        """Determines the weakly coupled disciplines, that is
        the disciplines that do not occur in MDAs."""
        weak_disciplines = []
        for parallel_tasks in self.sequence:
            for component in parallel_tasks:
                # find single disciplines
                if len(component) == 1 and not self.is_self_coupled(component[0]):
                    weak_disciplines.append(component[0])
        return weak_disciplines

    def strong_couplings(self):
        """Determines the strong couplings = outputs of the strongly
        coupled disciplines that are also inputs of the strongly
        coupled disciplines."""
        strong_disciplines = self.strongly_coupled_disciplines()
        # determine strong couplings = the outputs of the strongly coupled
        # disciplines that are inputs of any other discipline
        strong_couplings = set()
        for strong_discipline in strong_disciplines:
            strong_couplings.update(strong_discipline.get_output_data_names())
        inputs = set()
        for discipline in self.disciplines:
            inputs.update(discipline.get_input_data_names())
        return sorted(list(strong_couplings & inputs))

    def weak_couplings(self):
        """Determines the weak couplings = outputs of the weakly
        coupled disciplines."""
        weak_disciplines = self.weakly_coupled_disciplines()
        # determine strong couplings = the outputs of the weakly coupled
        # disciplines that are inputs of any other discipline
        weak_couplings = set()
        for weak_discipline in weak_disciplines:
            weak_couplings.update(weak_discipline.get_output_data_names())
        return sorted(list(weak_couplings))

    def input_couplings(self, discipline):
        """Computes all input coupling variables of a discipline.

        :param discipline: the discipline
        :returns: the list of input coupling variables names
        """
        input_names = discipline.get_input_data_names()
        strong_couplings = self.strong_couplings()
        return sorted([name for name in input_names if name in strong_couplings])

    def get_all_couplings(self):
        """
        Computes all coupling variables, weak or strong
        :returns: the list of coupling variables names
        """
        inputs = []
        outputs = []
        for discipline in self.disciplines:
            inputs += discipline.get_input_data_names()
            outputs += discipline.get_output_data_names()
        return sorted(list(set(inputs) & set(outputs)))

    def output_couplings(self, discipline, strong=True):
        """Computes the output coupling variables of a discipline.

        :param discipline: the discipline
        :returns: the list of output coupling variables names
        """
        output_names = discipline.get_output_data_names()
        if strong:
            couplings = self.strong_couplings()
        else:
            couplings = self.get_all_couplings()

        return sorted([name for name in output_names if name in couplings])

    def find_discipline(self, output):
        """Finds which discipline produces a given output.

        :param output: the name of the output
        :returns: the discipline if it is found, otherwise raise
            an exception

        """
        if not isinstance(output, string_types):
            raise TypeError("Output shall be a string")
        for discipline in self.disciplines:
            if discipline.is_output_existing(output):
                return discipline
        raise ValueError(output + " is not the output " + "of a discipline")

    def plot_n2_chart(
        self,
        file_path="n2.pdf",
        show_data_names=True,
        save=True,
        show=False,
        figsize=(15, 10),
    ):
        """
        Generates a N2 plot for the disciplines list.

        :param file_path: file path of the figure
        :param show_data_names: if true, the names of the
            coupling data is shown
            otherwise, circles are drawn, which size depend on the
            number of coupling names
        :param save: if True, saved the figure to file_path
        :param show: if True, shows the plot
        """

        fig = plt.figure(figsize=figsize)
        plt.rc("grid", linestyle="-", color="black", lw=1)
        plt.grid(True)
        axe = gca()
        n_disc = len(self.disciplines)
        ax_ticks = list(range(n_disc + 1))
        axe.xaxis.set_ticks(ax_ticks)
        axe.yaxis.set_ticks(ax_ticks)
        axe.xaxis.set_ticklabels([])
        axe.yaxis.set_ticklabels([])
        fig.tight_layout()

        for i, disc_i in enumerate(self.disciplines):
            x_1 = i
            x_2 = i + 1
            y_1 = n_disc - i
            y_2 = n_disc - i - 1
            plt.fill(
                [x_1, x_1, x_2, x_2], [y_1, y_2, y_2, y_1], "limegreen", alpha=0.45
            )
            text = plt.text(
                i + 0.5,
                n_disc - i - 0.5,
                disc_i.name,
                verticalalignment="center",
                horizontalalignment="center",
            )
            self._check_size_text(text, fig, n_disc)

        coupl_tuples = self.graph.get_disciplines_couplings()

        max_cpls = array([len(tpl[2]) for tpl in coupl_tuples]).max()

        for disc1, disc2, c_vars in coupl_tuples:
            i = self.disciplines.index(disc1)
            j = self.disciplines.index(disc2)
            if show_data_names:
                text = plt.text(
                    j + 0.5,
                    n_disc - i - 0.5,
                    "\n".join(c_vars),
                    verticalalignment="center",
                    horizontalalignment="center",
                )
                self._check_size_text(text, fig, n_disc)
            else:
                circle = plt.Circle(
                    (0.5 + j, n_disc - 0.5 - i),
                    len(c_vars) / (3.0 * max_cpls),
                    color="blue",
                )
                axe.add_artist(circle)

        if save:
            plt.savefig(file_path)
        if show:
            plt.show()

    @staticmethod
    def _check_size_text(text, figure, n_disc):
        """
        check the size of the text plotted in the N2 matrix and adapt
        the fig size according to the text shown

        :param text: text shown in the N2 matrix
        :param figure: figure of the n2 matrix
        :param n_disc: number of disciplines to be visible
        """
        renderer = figure.canvas.get_renderer()
        bbox = text.get_window_extent(renderer=renderer)
        width = bbox.width
        height = bbox.height
        size_max_box = figure.get_size_inches() * figure.dpi / n_disc
        inches = figure.get_size_inches()
        if width > size_max_box[0]:
            width_l = round(width * n_disc / figure.dpi) + 1
            figure.set_size_inches(width_l, inches[1])
        if height > size_max_box[1]:
            length_l = round(height * n_disc / figure.dpi) + 1
            figure.set_size_inches(inches[0], length_l)


class DependencyGraph(object):
    """Constructs a graph of dependency between the disciplines, and
    generate a sequence of execution (including strongly coupled
    disciplines, sequential tasks, parallel tasks)."""

    def __init__(self, disciplines):
        """
        Constructor

        :param disciplines: list of disciplines
        """
        self.disciplines = disciplines
        # initial graph
        initial_nodes = self._create_initial_nodes()
        (self.initial_graph, self.initial_edges) = self._compute_graph(initial_nodes)
        # strongly connected components
        self.components = self._strongly_connected_components()
        # resulting reduced graph
        reduced_nodes = self._create_component_nodes(initial_nodes)
        (self.reduced_graph, self.reduced_edges) = self._compute_graph(reduced_nodes)
        self.execution_sequence = self._topological_sort()

    def _create_initial_nodes(self):
        """Creates a list of (input, output) coupling variables
        for each discipline.
        The resulting list has the same order as the disciplines."""
        nodes = []
        for discipline in self.disciplines:
            nodes.append(
                (discipline.get_input_data_names(), discipline.get_output_data_names())
            )
        return nodes

    def _create_component_nodes(self, initial_nodes):
        """Creates a list of (input, output) coupling variables
        for each strongly connected component.
        The resulting list has the same order as the strongly connected
        components.

        :param initial_nodes: nodes of the initial graph
        """
        nodes = []
        for component in self.components:
            component_inputs = set()
            component_outputs = set()
            for node_index in component:
                (inputs, outputs) = initial_nodes[node_index]
                component_inputs.update(inputs)
                component_outputs.update(outputs)
            nodes.append((list(component_inputs), list(component_outputs)))
        return nodes

    @staticmethod
    def _compute_graph(nodes):
        """Computes the successors_i of each node and the edges between
        the nodes.

        :param nodes: the nodes of the graph
        """
        graph = {}
        disc_i = 0
        edges = {}

        for (_, outputs_i) in nodes:
            successors_i = set()
            # find out in which discipline(s) the outputs_i are used
            for output_i in outputs_i:
                disc_j = 0
                for (inputs_j, _) in nodes:
                    if disc_i != disc_j and output_i in inputs_j:
                        successors_i.add(disc_j)
                        # add the edge disc_i -> disc_j with
                        # label output_i
                        if disc_i not in edges:
                            edges[disc_i] = {}
                        if disc_j not in edges[disc_i]:
                            edges[disc_i][disc_j] = []
                        edges[disc_i][disc_j].append(output_i)
                    disc_j += 1

            graph[disc_i] = successors_i
            disc_i += 1
        return graph, edges

    def _strongly_connected_components(self):
        """Tarjan's algorithm determines the strongly connected components of a
        directed initial_graph.
        Within a component, there exists a path between each pair of nodes.

        Based on:
        http://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
        """
        stack = []
        comp_inds = {}
        node_indices = {}
        component_list = []  # result

        def strong_connect(node_index, current_index):
            """Browses the initial_graph from a node_index.

            :param node_index: current node index
            :param current_index:
            """
            node_indices[node_index] = current_index
            comp_inds[node_index] = current_index
            current_index += 1
            stack.append(node_index)

            # Consider successors of node_index
            successors = self.initial_graph.get(node_index, [])
            for successor in successors:
                if successor not in comp_inds:
                    # Successor has not yet been visited; recurse on it
                    current_index = strong_connect(successor, current_index)
                    comp_inds[node_index] = min(
                        comp_inds[node_index], comp_inds[successor]
                    )
                elif successor in stack:
                    # the successor is in the stack, therefore in the current
                    # strongly connected component
                    comp_inds[node_index] = min(
                        comp_inds[node_index], node_indices[successor]
                    )

            # If node_index is a root node_index, pop the stack and generate an
            # SCC
            if comp_inds[node_index] == node_indices[node_index]:
                # build the component by popping the stack
                connected_component = self._unstack_component(node_index, stack)
                component_list.append(tuple(connected_component))

            return current_index

        current_index = 0
        for node in self.initial_graph:
            if node not in comp_inds:
                current_index = strong_connect(node, current_index)

        return component_list

    @staticmethod
    def _unstack_component(node, stack):
        """Builds a set of nodes corresponding to a strongly connected
        component.

        :param node: current node
        :param stack: current stack
        """
        # build the component by popping the stack
        connected_component = []
        while True:
            successor = stack.pop()
            connected_component.append(successor)
            if successor == node:
                break
        return connected_component

    # TOPOLOGICAL SORT
    def _topological_sort(self):
        """Computes a topological sort of a directed graph.
        Determines if some nodes may be run in parallel."""
        current_graph = self.reduced_graph.copy()
        result = []
        while True:
            # find the leaves of the initial_graph
            leaves = set()
            for (node_index, successors) in current_graph.items():
                if not successors:
                    leaves.add(node_index)

            if not leaves:
                break
            # all leaves are parallelizable
            parallel_tasks = set(
                self.components[component_index] for component_index in leaves
            )
            result.append(sorted(parallel_tasks))
            # update the current initial_graph by removing the leaves and
            # the initial_edges pointing toward them
            current_graph = {
                node_index: (successors - leaves)
                for (node_index, successors) in current_graph.items()
                if node_index not in leaves
            }
        return result[::-1]

    def get_disciplines_couplings(self):
        """Returns couplings between disciplines as a list of
        3-uples (from_disc, to_disc, variables names set).
        """
        couplings = []
        for from_disc in self.initial_edges:
            for to_disc in self.initial_edges[from_disc]:
                couplings.append(
                    (
                        self.disciplines[from_disc],
                        self.disciplines[to_disc],
                        sorted(self.initial_edges[from_disc][to_disc]),
                    )
                )
        return couplings

    # EXPORT METHODS
    def export_initial_graph(self, file_path):
        """Exports a visualization of the initial graph.

        :param file_path: file path of the generated file
        """
        file_name, file_extension = os.path.splitext(file_path)
        dot = Digraph(comment="Dependency graph", format=file_extension[1:])
        # add the disciplines as nodes
        disc_dict = {}
        for i, disc in enumerate(self.disciplines):
            disc_id = str(i)
            dot.node(disc_id, disc.name)
            disc_dict[disc] = disc_id

        # add the coupling as initial_edges
        for disc_from in self.initial_edges:
            for disc_to in self.initial_edges[disc_from]:
                outputs = self.initial_edges[disc_from][disc_to]
                dot.edge(str(disc_from), str(disc_to), label=",".join(sorted(outputs)))
                # outputs of the last node

        last_tasks = self.execution_sequence[-1]
        for k, last_task in enumerate(last_tasks):
            last_taskind = last_task[0]
            i_str = "-" + str(k)
            disc = self.disciplines[last_taskind]
            last_outputs = disc.get_output_data_names()
            if last_outputs != []:
                # create an edge to an invisible node
                dot.node(i_str, style="invis", shape="point")
                label = ",".join(last_outputs)
                dot.edge(disc_dict[disc], i_str, label=label)

        dot.render(file_name, view=False)
        move(file_name, "{}.gv".format(file_name))

    def export_reduced_graph(self, file_path):
        """Exports a visualization of the reduced graph.

        :param file_path: the file_path of the generated file
        """
        file_name, file_extension = os.path.splitext(file_path)
        dot = Digraph(comment="Dependency graph", format=file_extension[1:])
        # add the disciplines as nodes
        for parallel_tasks in self.execution_sequence:
            for component in parallel_tasks:
                # if MDA, aggregate the names of the disciplines
                if len(component) > 1:
                    disc_names = [self.disciplines[disc].name for disc in component]
                    component_name = "MDA of " + ", ".join(disc_names)
                else:
                    component_name = self.disciplines[component[0]].name
                dot.node(str(component), component_name)

        # outputs of the last nodes

        last_tasks = self.execution_sequence[-1]
        for i, last_task in enumerate(last_tasks):
            last_taskind = last_task[0]
            i_str = "-" + str(i)
            last_outputs = self.disciplines[last_taskind].get_output_data_names()
            if last_outputs != []:
                # create an edge to an invisible node
                dot.node(i_str, style="invis", shape="point")
                label = ",".join(last_outputs)
                dot.edge(str(last_task), i_str, label=label)

        # add the coupling as initial_edges
        for disc_from in self.reduced_edges:
            for disc_to in self.reduced_edges[disc_from]:
                outputs = self.reduced_edges[disc_from][disc_to]
                dot.edge(
                    str(self.components[disc_from]),
                    str(self.components[disc_to]),
                    label=",".join(sorted(outputs)),
                )
        dot.render(file_name, view=False)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Regression model
================

The :mod:`~gemseo.mlearning.regression.regression` module
implements regression algorithms, where the goal is to find relationships
between continuous input and output variables. After being fitted to a learning
set, Regression algorithms can predict output values of new input data.

A regression algorithm consists of identifying a function
:math:`f: \\mathbb{R}^{n_{\\textrm{inputs}}} \\to
\\mathbb{R}^{n_{\\textrm{outputs}}}`. Given an input point
:math:`x \\in \\mathbb{R}^{n_{\\textrm{inputs}}}`, the predict method of the
regression algorithm will return the output point
:math:`y = f(x) \\in \\mathbb{R}^{n_{\\textrm{outputs}}}`. See
:mod:`~gemseo.mlearning.core.supervised` for more information.

Wherever possible, regression algorithms should also be able to compute the
Jacobian matrix of the function it has learned to represent. Given an input
point :math:`x \\in \\mathbb{R}^{n_{\\textrm{inputs}}}`, the Jacobian predict
method of the regression algorithm should thus return the matrix

.. math::

    J_f(x) = \\frac{\\partial f}{\\partial x} =
    \\begin{pmatrix}
    \\frac{\\partial f_1}{\\partial x_1} & \\cdots & \\frac{\\partial f_1}
        {\\partial x_{n_{\\textrm{inputs}}}}\\\\
    \\vdots & \\ddots & \\vdots\\\\
    \\frac{\\partial f_{n_{\\textrm{outputs}}}}{\\partial x_1} & \\cdots &
        \\frac{\\partial f_{n_{\\textrm{outputs}}}}
        {\\partial x_{n_{\\textrm{inputs}}}}
    \\end{pmatrix}
    \\in \\mathbb{R}^{n_{\\textrm{outputs}}\\times n_{\\textrm{inputs}}}.

This concept is implemented through
the :class:`.MLRegressionAlgo` class which
inherits from the :class:`.MLSupervisedAlgo` class.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import eye, matmul

from gemseo.mlearning.core.supervised import MLSupervisedAlgo
from gemseo.utils.data_conversion import DataConversion

standard_library.install_aliases()


class MLRegressionAlgo(MLSupervisedAlgo):
    """Machine Learning Regression Model Algorithm.

    Inheriting classes should implement the :meth:`!MLSupervisedAlgo._fit` and
    :meth:`!MLSupervisedAlgo._predict` methods, and
    :meth:`!MLRegressionAlgo._predict_jacobian` method if possible.
    """

    class DataFormatters(MLSupervisedAlgo.DataFormatters):
        """ Machine learning regression model decorators. """

        @classmethod
        def format_dict_jacobian(cls, predict):
            """If input_data is passed as a dictionary, then convert it to
            ndarray, and convert output_data to dictionary. Else, do nothing.

            :param predict: Method whose input_data and output_data are to be
                formatted.
            """

            def wrapper(self, input_data, *args, **kwargs):
                as_dict = isinstance(input_data, dict)
                if as_dict:
                    input_data = DataConversion.dict_to_array(
                        input_data, self.input_names
                    )
                single_sample = len(input_data.shape) == 1
                jacobians = predict(self, input_data, *args, **kwargs)
                if as_dict:
                    varsizes = self.learning_set.sizes
                    if single_sample:
                        jacobians = DataConversion.jac_2dmat_to_dict(
                            jacobians, self.output_names, self.input_names, varsizes
                        )
                    else:
                        jacobians = DataConversion.jac_3dmat_to_dict(
                            jacobians, self.output_names, self.input_names, varsizes
                        )
                return jacobians

            return wrapper

        @classmethod
        def transform_jacobian(cls, predict_jac):
            """Apply transform to inputs, and inverse transform to outputs.

            :param predict: Method whose input_data and output_data are to be
                formatted.
            """

            def wrapper(self, input_data, *args, **kwargs):
                inputs = self.learning_set.INPUT_GROUP
                if inputs in self.transformer:
                    jac = self.transformer[inputs].compute_jacobian(input_data)
                    input_data = self.transformer[inputs].transform(input_data)
                else:
                    jac = eye(input_data.shape[1])

                jac = matmul(predict_jac(self, input_data, *args, **kwargs), jac)
                output_data = self.predict_raw(input_data)

                outputs = self.learning_set.OUTPUT_GROUP
                if outputs in self.transformer:
                    jac = matmul(
                        self.transformer[outputs].compute_jacobian_inverse(output_data),
                        jac,
                    )
                return jac

            return wrapper

    def predict_raw(self, input_data):
        """Predict output data from input data, assuming both are 2D.

        :param ndarray input_data: input data (n_samples, n_inputs).
        :return: output data (n_samples, n_outputs).
        :rtype: ndarray(int)
        """
        return self._predict(input_data)

    @DataFormatters.format_dict_jacobian
    @DataFormatters.format_samples
    @DataFormatters.transform_jacobian
    def predict_jacobian(self, input_data):
        """Predict Jacobian of the regression model of input_data.

        :param input_data: 1D input data.
        :type input_data: dict(ndarray) or ndarray
        :return: Jacobian for given input data.
        :rtype: dict(dict(ndarray)) or ndarray
        """
        return self._predict_jacobian(input_data)

    def _predict_jacobian(self, input_data):
        """Predict Jacobian of the regression model for the given input data.

        :param ndarray input_data: input_data (2D).
        :return: Jacobian matrices (3D, one for each sample).
        :rtype: ndarray
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""
RBF regression
==============

The radial basis function surrogate discipline expresses the model output
as a weighted sum of kernel functions centered on the learning input data:

.. math::

    y = w_1K(\|x-x_1\|;\epsilon) + w_2K(\|x-x_2\|;\epsilon) + ...
        + w_nK(\|x-x_n\|;\epsilon)

and the coefficients :math:`(w_1, w_2, ..., w_n)`
are estimated by least square regression.

Dependence
----------
The RBF model relies on the Rbf class
of the `scipy library <https://docs.scipy.org/doc/scipy/reference/generated/
scipy.interpolate.Rbf.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

import pickle
from os.path import join

from future import standard_library
from numpy import array, average, exp, finfo, hstack, log, sqrt
from numpy.linalg import norm
from scipy.interpolate import Rbf
from six import string_types

from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.utils.py23_compat import PY3

standard_library.install_aliases()


from gemseo import LOGGER


class RBFRegression(MLRegressionAlgo):
    """ Regression based on radial basis functions. """

    LIBRARY = "scipy"
    ABBR = "RBF"

    EUCLIDEAN = "euclidean"

    MULTIQUADRIC = "multiquadric"
    INVERSE_MULTIQUADRIC = "inverse_multiquadric"
    GAUSSIAN = "gaussian"
    LINEAR = "linear"
    CUBIC = "cubic"
    QUINTIC = "quintic"
    THIN_PLATE = "thin_plate"

    AVAILABLE_FUNCTIONS = [
        MULTIQUADRIC,
        INVERSE_MULTIQUADRIC,
        GAUSSIAN,
        LINEAR,
        CUBIC,
        QUINTIC,
        THIN_PLATE,
    ]

    def __init__(
        self,
        data,
        transformer=None,
        input_names=None,
        output_names=None,
        function=MULTIQUADRIC,
        der_function=None,
        epsilon=None,
        **parameters
    ):
        r"""Constructor.

        :param data: learning dataset
        :type data: Dataset
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables. Default: None.
        :type input_names: list(str)
        :param output_names: names of the output variables. Default: None.
        :type output_names: list(str)
        :param function: radial basis function. Default: 'multiquadric'.
        :type function: str or callable
        :param der_function: derivative of radial basis function, only to be
            provided if function is callable and not str. The der_function
            should take three arguments (input_data, norm_input_data, eps). For
            a RBF of the form function(:math:`r`),
            der_function(:math:`x`, :math:`|x|`, :math:`\epsilon`) should
            return :math:`\epsilon^{-1} x/|x| f'(|x|/\epsilon)`. Default: None.
        :type der_function: callable
        :param epsilon: Adjustable constant for Gaussian or
            multiquadrics functions. Default: None.
        :type epsilon: float
        :param parameters: other RBF parameters (sklearn).
        """
        if isinstance(function, string_types):
            function = str(function)
        super(RBFRegression, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            function=function,
            epsilon=epsilon,
            **parameters
        )
        self.y_average = 0.0
        self.der_function = der_function

    class RBFDerivatives(object):
        r"""Derivatives of functions used in RBFRegression.

        For a RBF of the form :math:`f(r)`, :math:`r` scalar,
        the derivative functions are defined by :math:`d(f(r))/dx`,
        with :math:`r=|x|/\epsilon`. The functions are thus defined
        by :math:`df/dx = \epsilon^{-1} x/|x| f'(|x|/\epsilon)`.
        This convention is chosen to avoid division by :math:`|x|` when
        the terms may be cancelled out, as :math:`f'(r)` often has a term
        in :math:`r`.
        """

        TOL = finfo(float).eps

        @classmethod
        def der_multiquadric(cls, input_data, norm_input_data, eps):
            r"""Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = \sqrt{r^2 + 1}`.

            :param float input_data: Input variable (vector).
            :param float norm_input_data: Norm of input variable.
            :return: Derivative of the function.
            :rtype: float
            """
            return input_data / eps ** 2 / sqrt((norm_input_data / eps) ** 2 + 1)

        @classmethod
        def der_inverse_multiquadric(cls, input_data, norm_input_data, eps):
            r"""Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = 1/\sqrt{r^2 + 1}`.

            :param float input_data: Input variable (vector).
            :param float norm_input_data: Norm of input variable.
            :return: Derivative of the function.
            :rtype: float
            """
            return -input_data / eps ** 2 / ((norm_input_data / eps) ** 2 + 1) ** 1.5

        @classmethod
        def der_gaussian(cls, input_data, norm_input_data, eps):
            r"""Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = \exp(-r^2)`.

            :param float input_data: Input variable (vector).
            :param float norm_input_data: Norm of input variable.
            :return: Derivative of the function.
            :rtype: float
            """
            return -2 * input_data / eps ** 2 * exp(-((norm_input_data / eps) ** 2))

        @classmethod
        def der_linear(cls, input_data, norm_input_data, eps):
            """Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = r`.
            If :math:`x=0`, return 0 (determined up to a tolerance).

            :param float input_data: Input variable (vector).
            :param float norm_input_data: Norm of input variable.
            :return: Derivative of the function.
            :rtype: float
            """
            return (
                (norm_input_data > cls.TOL)
                * input_data
                / eps
                / (norm_input_data + cls.TOL)
            )

        @classmethod
        def der_cubic(cls, input_data, norm_input_data, eps):
            """Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = r^3`.

            :param float input_data: Input variable (vector) :math:`x`.
            :param float norm_input_data: Norm of input variable :math:`|x|`.
            :return: Derivative of the function.
            :rtype: float
            """
            return 3 * norm_input_data * input_data / eps ** 3

        @classmethod
        def der_quintic(cls, input_data, norm_input_data, eps):
            """Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = r^5`.

            :param float input_data: Input variable (vector).
            :param float norm_input_data: Norm of input variable.
            :return: Derivative of the function.
            :rtype: float
            """
            return 5 * norm_input_data ** 3 * input_data / eps ** 5

        @classmethod
        def der_thin_plate(cls, input_data, norm_input_data, eps):
            r"""Compute derivative w.r.t. :math:`x` of the function
            :math:`f(r) = r^2 \log(r)`.
            If :math:`x=0`, return 0 (determined up to a tolerance).

            :param float input_data: Input variable (vector).
            :param float norm_input_data: Norm of input variable.
            :return: Derivative of the function.
            :rtype: float
            """
            return (
                (norm_input_data > cls.TOL)
                * input_data
                / eps ** 2
                * (1 + 2 * log(norm_input_data / eps + cls.TOL))
            )

    def _fit(self, input_data, output_data):
        """Fit the regression model.

        :param ndarray input_data: input data (2D)
        :param ndarray output_data: output data (2D)
        """
        self.y_average = average(output_data, axis=0)
        output_data -= self.y_average
        if PY3:
            args = list(input_data.T) + [output_data]
            self.algo = Rbf(*args, mode="N-D", smooth=0.0, **self.parameters)
        else:
            self.algo = []
            for output in range(output_data.shape[1]):
                args = hstack([input_data, output_data[:, [output]]])
                rbf = Rbf(*args.T, smooth=0.0, **self.parameters)
                self.algo.append(rbf)

    def _predict(self, input_data):
        """Predict output for given input data.

        :param ndarray input_data: input data (2D).
        :return: output prediction (2D).
        :rtype: ndarray
        """
        if PY3:
            output_data = self.algo(*input_data.T)
            if len(output_data.shape) == 1:
                output_data = output_data[:, None]  # n_outputs=1, rbf reduces
            output_data = output_data + self.y_average
        else:
            output_data = [rbf(*input_data.T) for rbf in self.algo]
            output_data = array(output_data).T + self.y_average
        return output_data

    def _predict_jacobian(self, input_data):
        """Predict Jacobian of the regression model for the given input data.

        :param ndarray input_data: input_data (2D).
        :return: Jacobian matrices (3D, one for each sample).
        :rtype: ndarray
        """
        self._check_available_jacobian()
        der_func = self.der_function or getattr(
            self.RBFDerivatives, "der_{}".format(self.function)
        )
        #             predict_samples                        learn_samples
        # Dimensions : ( n_samples , n_outputs , n_inputs , n_learn_samples )
        # input_data : ( n_samples ,           , n_inputs ,                 )
        # ref_points : (           ,           , n_inputs , n_learn_samples )
        # nodes      : (           , n_outputs ,          , n_learn_samples )
        # jacobians  : ( n_samples , n_outputs , n_inputs ,                 )
        if PY3:
            eps = self.algo.epsilon
            ref_points = self.algo.xi[None, None]
            nodes = self.algo.nodes.T[None, :, None]
        else:
            eps = [rbf.epsilon for rbf in self.algo]
            eps = array(eps)[None, :, None, None]  # 1 epsilon for each output
            ref_points = self.algo[0].xi  # same xi for all algos
            nodes = array([rbf.nodes for rbf in self.algo])[None, :, None]
        input_data = input_data[:, None, :, None]
        diffs = input_data - ref_points
        dists = norm(diffs, axis=2)[:, :, None]
        contributions = nodes * der_func(diffs, dists, eps=eps)
        jacobians = contributions.sum(-1)
        return jacobians

    def _check_available_jacobian(self):
        """ Check if the Jacobian is available for the given setup. """
        if PY3:
            norm_name = self.algo.norm
        else:
            norm_name = self.algo[0].norm

        if norm_name != self.EUCLIDEAN:
            raise NotImplementedError(
                "Jacobian is only implemented for " "euclidean norm."
            )

        if callable(self.function) and self.der_function is None:
            raise NotImplementedError(
                "No der_function is provided. Add "
                "der_function in RBFRegression "
                "constructor."
            )

    def _save_algo(self, directory):
        """Save external machine learning algorithm.

        :param str directory: algorithm directory.
        """
        if PY3:
            super(RBFRegression, self)._save_algo(directory)
        else:
            filename = join(directory, "algo.pkl")
            with open(filename, "wb") as handle:
                pickled_rbf = pickle.Pickler(handle, protocol=2)
                pickled_rbf_list = []
                for rbf in self.algo:
                    pickled_rbf_list.append({})
                    for key in rbf.__dict__.keys():
                        if key not in ["_function", "norm"]:
                            pickled_rbf_list[-1][key] = rbf.__getattribute__(key)
                pickled_rbf.dump(pickled_rbf_list)

    def load_algo(self, directory):
        """Load external machine learning algorithm.

        :param str directory: algorithm directory.
        """
        if PY3:
            super(RBFRegression, self).load_algo(directory)
        else:
            filename = join(directory, "algo.pkl")
            self.algo = []
            with open(filename, "rb") as handle:
                unpickled_rbf = pickle.Unpickler(handle)
                unpickled_rbf_list = unpickled_rbf.load()
                for rbf in unpickled_rbf_list:
                    algo_i = Rbf(
                        array([1, 2, 3]),
                        array([10, 20, 30]),
                        array([100, 200, 300]),
                        function=rbf["function"],
                    )
                    for key, value in rbf.items():
                        algo_i.__setattr__(key, value)
                    self.algo.append(algo_i)

    def _get_objects_to_save(self):
        """ Get objects to save. """
        objects = super(RBFRegression, self)._get_objects_to_save()
        objects["y_average"] = self.y_average
        objects["der_function"] = self.der_function
        return objects

    @property
    def function(self):
        """Kernel function name.

        The name is possibly different from self.parameters['function'], as it
        is mapped (scipy). Examples:

        'inverse'              -> 'inverse_multiquadric'
        'InverSE MULtiQuadRIC' -> 'inverse_multiquadric'

        return: Name of kernel function.
        rtype: str
        """
        if PY3:
            function = self.algo.function
        else:
            function = self.algo[0].function
        return function

    @classmethod
    def get_available_functions(cls):
        """ Get available RBFs. """
        return cls.AVAILABLE_FUNCTIONS



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Analytic MDODiscipline based on symbolic expressions
****************************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from future import standard_library
from numpy import array, float64, zeros
from six import string_types
from sympy.parsing.sympy_parser import parse_expr

from gemseo.core.discipline import MDODiscipline

standard_library.install_aliases()


from gemseo import LOGGER


class AnalyticDiscipline(MDODiscipline):
    """Discipline based on analytic expressions list,
    using the symbolic calculation sympy engine.

    Automatically differentiates the expressions to obtain
    the Jacobian matrices.

    See also
    --------
    gemseo.core.discipline.MDODiscipline : abstract class defining
        the key concept of discipline
    """

    def __init__(self, name=None, expressions_dict=None):
        """
        Constructor

        :param name: name of the discipline.
        :param expressions_dict: dictionary of outputs and their expressions
            for instance : { 'y_1':'2*x**2', 'y_2':'4*x**2+5+z**3'}
            will create a discipline with outputs y_1, y_2
            and inputs x, and z.
        """
        super(AnalyticDiscipline, self).__init__(name)
        if not expressions_dict:
            raise ValueError("expressions_dict is a mandatory argument")
        self.expressions_dict = expressions_dict
        self.expr_symbols_dict = {}
        self.input_names = []
        self._sympy_exprs = {}
        self._sympy_jac_exprs = {}
        self._init_expressions()
        self._init_grammars()
        self._init_default_inputs()
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _init_grammars(self):
        """Initializes the |g| grammars from the expressions dict"""
        zero = zeros(2)
        in_dict = {k: zero for k in self.input_names}
        self.input_grammar.initialize_from_base_dict(in_dict)
        out_dict = {k: zero for k in self.expressions_dict}
        self.output_grammar.initialize_from_base_dict(out_dict)

    def _init_expressions(self):
        """Parses the expressions and get sympy exprs
        idem for jacobians


        """

        input_symbols = []
        for out_var, expr in self.expressions_dict.items():
            if not isinstance(expr, string_types):
                raise TypeError("Expressions must be an iterable of strings")
            parsed = parse_expr(expr)
            args = list(parsed.free_symbols)
            self._sympy_exprs[out_var] = parsed
            input_symbols += args
            args_names = [arg.name for arg in args]
            self.expr_symbols_dict[out_var] = args_names
            self._sympy_jac_exprs[out_var] = {}
            for arg in args_names:
                jac_expr = parsed.diff(arg)
                self._sympy_jac_exprs[out_var][arg] = jac_expr

        self.input_names = [symb.name for symb in set(input_symbols)]

    def _init_default_inputs(self):
        """Initalizes the default inputs of the discipline
        with zeros


        """
        zeros_array = zeros(1)
        self.default_inputs = {k: zeros_array for k in self.input_names}

    def _run(self):
        """
        Runs the discipline
        """
        outputs = {}
        # Do not pass useless tokens to the expr, this may
        # fail when tokens contains dots, or slow down the process
        filtered_inputs = {key: float(val.real) for key, val in self.local_data.items()}
        for out_var, expr in self._sympy_exprs.items():
            try:
                out_val = expr.evalf(subs=filtered_inputs)
                outputs[out_var] = array([out_val], dtype=float64)
            except TypeError:
                LOGGER.error("Failed to evaluate expression : %s", str(expr))
                LOGGER.error("With inputs : %s", str(self.local_data))
                raise
        self.store_local_data(**outputs)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """
        Computes the jacobian

        :param inputs: Default value = None)
        :param outputs: Default value = None)
        """
        # otherwise there may be missing terms
        # if some formula have no dependency
        self._init_jacobian(inputs, outputs, with_zeros=True)
        filtered_inputs = {key: float(val.real) for key, val in self.local_data.items()}
        for out_var, expr in self._sympy_exprs.items():
            for arg in expr.free_symbols:
                j_expr = self._sympy_jac_exprs[out_var][arg.name]
                jac_val = j_expr.evalf(subs=filtered_inputs)
                self.jac[out_var][arg.name] = array([[jac_val]], dtype=float64)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                  initial documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable model
==============

This module implements the abstract concept of scalable model
which is used by scalable disciplines. A scalable model is built
from a input-output learning dataset associated with a function
and generalizing its behavior to a new user-defined problem dimension,
that is to say new used-defined input and output dimensions.

The concept of scalable model is implemented
through :class:`.ScalableModel`, an abstract class which is instantiated from:

- data provided as an :class:`.AbstractFullCache`
- variables sizes provided as a dictionary
  whose keys are the names of inputs and outputs
  and values are their new sizes.
  If a variable is missing, its original size is considered.

Scalable model parameters can also be filled in.
Otherwise the model uses default values.

.. seealso::

   The :class:`.ScalableDiagonalModel` class overloads :class:`.ScalableModel`.
"""
from __future__ import absolute_import, division, unicode_literals

from copy import deepcopy

from future import standard_library
from numpy import maximum as np_max
from numpy import minimum as np_min
from numpy import ones
from past.utils import old_div

from gemseo.caches.memory_full_cache import MemoryFullCache

standard_library.install_aliases()

from gemseo import LOGGER


class ScalableModel(object):
    """ Scalable model. """

    ABBR = "sm"

    def __init__(self, data, sizes=None, **parameters):
        """Constructor.

        :param AbstractFullCache data: learning dataset.
        :param dict sizes: sizes of input and output variables.
            If None, use the original sizes.
            Default: None.
        :param parameters: model parameters
        """
        sizes = sizes or {}
        self.name = self.ABBR + "_" + data.name
        self.data = data
        self.sizes = self._set_sizes(sizes)
        self.parameters = parameters
        self.lower_bounds, self.upper_bounds = self.compute_bounds()
        self.normalize_data()
        self.lower_bounds, self.upper_bounds = self.compute_bounds()
        self.default_inputs = self._set_default_inputs()
        self.model = self.build_model()

    def _set_default_inputs(self):
        """Sets the default values of inputs from the model.

        :return: default inputs.
        :rtype: dict
        """
        default_inputs = {}
        for name in self.inputs_names:
            default_inputs[name] = 0.5 * ones(self.sizes[name])
        return default_inputs

    def scalable_function(self, input_value=None):
        """Evaluate the scalable function.

        :param dict input_value: input values.
            If None, use default inputs. Default: None.
        :return: evaluation of the scalable function.
        :rtype: dict
        """
        raise NotImplementedError

    def scalable_derivatives(self, input_value=None):
        """Evaluate the scalable derivatives.

        :param dict input_value: input values.
            If None, use default inputs. Default: None
        :return: evaluation of the scalable derivatives.
        :rtype: dict
        """
        raise NotImplementedError

    def compute_bounds(self):
        """Compute lower and upper bounds of both input and output variables.

        :return: lower bounds, upper bounds.
        :rtype: dict, dict
        """
        all_data = self.data.get_all_data(True)
        data = next(all_data)
        var_lb = deepcopy(data[self.data.INPUTS_GROUP])
        var_ub = deepcopy(data[self.data.INPUTS_GROUP])
        var_lb.update(data[self.data.OUTPUTS_GROUP])
        var_ub.update(data[self.data.OUTPUTS_GROUP])
        for data in all_data:
            input_data = data[self.data.INPUTS_GROUP]
            for varname, value in input_data.items():
                var_lb[varname] = np_min(var_lb[varname], value)
                var_ub[varname] = np_max(var_ub[varname], value)
            output_data = data[self.data.OUTPUTS_GROUP]
            for varname, value in output_data.items():
                var_lb[varname] = np_min(var_lb[varname], value)
                var_ub[varname] = np_max(var_ub[varname], value)
        all_data = self.data.get_all_data(True)
        return var_lb, var_ub

    def normalize_data(self):
        """ Normalize cache from lower and upper bounds. """

        def normalize_vector(value, name):
            """ Normalize vector. """
            lower = self.lower_bounds[name]
            upper = self.upper_bounds[name]
            return old_div((value - lower), (upper - lower))

        normalized_cache = MemoryFullCache()
        all_data = self.data.get_all_data(True)
        for data in all_data:
            input_data = data[self.data.INPUTS_GROUP]
            input_data = {
                name: normalize_vector(value, name)
                for name, value in input_data.items()
            }
            output_data = data[self.data.OUTPUTS_GROUP]
            output_data = {
                name: normalize_vector(value, name)
                for name, value in output_data.items()
            }
            normalized_cache.cache_outputs(
                input_data, self.inputs_names, output_data, self.outputs_names
            )
        self.data = normalized_cache

    def build_model(self):
        """ Build model with original sizes for input and output variables."""
        raise NotImplementedError

    @property
    def original_sizes(self):
        """Original sizes of variables.

        :return: original sizes of variables.
        :rtype: dict
        """
        return self.data.varsizes

    @property
    def outputs_names(self):
        """Outputs names.

        :return: names of the outputs.
        :rtype: list(str)
        """
        return sorted(self.data.outputs_names)

    @property
    def inputs_names(self):
        """Inputs names.

        :return: names of the inputs.
        :rtype: list(str)
        """
        return sorted(self.data.inputs_names)

    def _set_sizes(self, sizes):
        """Set the new sizes of input and output variables.

        :param dict: new sizes of some variables.
        :return: new sizes of all variables.
        :rtype: dict
        """
        for name in self.data.inputs_names + self.data.outputs_names:
            original_size = self.original_sizes.get(name)
            sizes[name] = sizes.get(name, original_size)
        return sizes



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Composed distribution based on OpenTURNS
========================================

Overview
--------

The :class:`.OTComposedDistribution` class is a concrete class
inheriting from :class:`.ComposedDistribution` which is an abstract one.
OT stands for `OpenTURNS <http://www.openturns.org/>`_
which is the library it relies on.

Construction
------------

The :class:`.OTComposedDistribution` of a list of given uncertain variables
is built from a list of :class:`.OTDistribution` objects
implementing the probability distributions of these variables
based on the OpenTURNS library and from a copula name.

.. note::

   A copula is a mathematical function used to define the dependence
   between random variables from their cumulative density functions.
   `See more <https://en.wikipedia.org/wiki/Copula_(probability_theory)>`__.
"""
from __future__ import absolute_import, division, unicode_literals

import openturns as ots
from future import standard_library
from numpy import array

from gemseo.uncertainty.distributions.distribution import ComposedDistribution

standard_library.install_aliases()


class OTComposedDistribution(ComposedDistribution):
    """ OpenTURNS composed distribution. """

    COPULA = {ComposedDistribution.INDEPENDENT_COPULA: ots.IndependentCopula}

    def __init__(self, distributions, copula=ComposedDistribution.INDEPENDENT_COPULA):
        """Constructor.

        :param list(OTDistribution) distributions: list of OT distributions.
        :param str copula: copula name.
        """
        super(OTComposedDistribution, self).__init__(distributions, copula)
        marginals = [
            marginal
            for distribution in distributions
            for marginal in distribution.marginals
        ]
        ot_copula = self.COPULA[copula](len(marginals))
        self.distribution = ots.ComposedDistribution(marginals, ot_copula)
        self._mapping = {}
        index = 0
        for id_dist, distribution in enumerate(distributions):
            for id_marg in range(distribution.dimension):
                self._mapping[index] = (id_dist, id_marg)
                index += 1
        self._set_bounds(distributions)

    def get_sample(self, n_samples=1):
        """Get sample.

        :param n_samples: number of samples.
        :type n_samples: int
        :return: samples
        :rtype: list(array)
        """
        sample = array(self.distribution.getSample(n_samples))
        return sample

    def cdf(self, vector):
        """Evaluate the cumulative density functions of the marginals
        of a random variable for a given instance.

        :param array vector: instance of the random variable.
        """
        tmp = []
        for index, value in enumerate(vector):
            id1 = self._mapping[index][0]
            id2 = self._mapping[index][1]
            value = ots.Point([value])
            tmp.append(self.marginals[id1].marginals[id2].computeCDF(value))
        return array(tmp)

    def inverse_cdf(self, vector):
        """Evaluate the inverses of the cumulative density functions of the
        marginals of a random variable for a given vector .

        :param array vector: vector of values comprised between 0 and 1 with
            same dimension as the random variable.
        """
        tmp = []
        for index, value in enumerate(vector):
            id1 = self._mapping[index][0]
            id2 = self._mapping[index][1]
            tmp.append(self.marginals[id1].marginals[id2].computeQuantile(value)[0])
        return array(tmp)

    def _pdf(self, index):
        id1 = self._mapping[index][0]
        id2 = self._mapping[index][1]

        def pdf(level):
            return self.marginals[id1].marginals[id2].computePDF(level)

        return pdf

    def _cdf(self, index):
        id1 = self._mapping[index][0]
        id2 = self._mapping[index][1]

        def cdf(level):
            return self.marginals[id1].marginals[id2].computeCDF(level)

        return cdf



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""
Sobol' indices
==============
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import array
from openturns import (
    JansenSensitivityAlgorithm,
    MartinezSensitivityAlgorithm,
    MauntzKucherenkoSensitivityAlgorithm,
    SaltelliSensitivityAlgorithm,
    Sample,
)

from gemseo.core.discipline import MDODiscipline
from gemseo.core.doe_scenario import DOEScenario
from gemseo.utils.data_conversion import DataConversion

standard_library.install_aliases()

from gemseo import LOGGER


class SobolIndices(object):
    """ Sobol' indices. """

    ALGOS = {
        "Saltelli": SaltelliSensitivityAlgorithm,
        "Jansen": JansenSensitivityAlgorithm,
        "MauntzKucherenko": MauntzKucherenkoSensitivityAlgorithm,
        "Martinez": MartinezSensitivityAlgorithm,
    }

    def __init__(
        self, disciplines, space, n_samples, formulation="MDF", objective_name=None
    ):
        """Constructor

        :param list(MDODiscipline) disciplines: disciplines.
        :param ParameterSpace space: parameter space.
        :param int n_samples: number of samples.
        :param str formulation: MDO formulation. Default: 'MDF'.
        :param str objective_name: objective name. If None, use the first
            output. Default: None.
        """
        if isinstance(disciplines, MDODiscipline):
            disciplines = [disciplines]
        first_output = next(iter(disciplines[0].get_output_data_names()))
        objective_name = objective_name or first_output
        scenario = DOEScenario(disciplines, formulation, objective_name, space)
        scenario.execute({"algo": "OT_SOBOL_INDICES", "n_samples": n_samples})
        opt_problem = scenario.formulation.opt_problem
        self.dataset = opt_problem.export_to_dataset("sobol", opt_naming=False)
        self.n_samples = len(self.dataset)

    def get_indices(self, algo="Saltelli"):
        """Get Sobol' indices.

        :param str algo: method to compute the Sobol' indices,
            either 'Saltelli', 'Jansen', 'MauntzKucherenko' or 'Martinez'.
            Default: 'Saltelli'.
        """
        try:
            algo = self.ALGOS[algo]
        except Exception:
            raise TypeError(
                algo + " is not an available algorithm " "to compute Sobol" " indices."
            )
        array_to_dict = DataConversion.array_to_dict
        inputs_names = self.dataset.get_names("inputs")
        sizes = self.dataset.sizes
        inputs = Sample(self.dataset.get_data_by_group("inputs"))
        outputs = Sample(self.dataset.get_data_by_group("outputs"))
        dim = self.dataset.dimension[self.dataset.INPUT_GROUP]
        n_samples = int(self.n_samples / (dim + 2))
        sobol = algo(inputs, outputs, n_samples)
        first_order = array(sobol.getFirstOrderIndices())
        first_order = array_to_dict(first_order, inputs_names, sizes)
        total_order = array(sobol.getTotalOrderIndices())
        total_order = array_to_dict(total_order, inputs_names, sizes)
        return sobol, first_order, total_order



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Pierre-Jean Barjhoux
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A constraints plot
******************
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from matplotlib import pyplot as plt
from numpy import arange, array

from gemseo.post.opt_post_processor import OptPostProcessor

standard_library.install_aliases()


from gemseo import LOGGER


class BasicHistory(OptPostProcessor):
    """
    The **BasicHistory** post processing plots any of the constraint
    or objective functions
    w.r.t. optimization iterations or sampling snapshots.

    The plot method requires the list of variable names to plot.
    It is possible either to save the plot, to show the plot or both.
    """

    def _plot(
        self,
        data_list,
        show=False,
        save=False,
        file_path="basic_history",
        extension="pdf",
    ):
        """
        Plots the optimization history:
        1 plot for the constraints

        :param data_list: list of variable names
        :type data_list: list(str)
        :param show: if True, displays the plot windows
        :type show: bool
        :param save: if True, exports plot to pdf
        :type save: bool
        :param file_path: the base paths of the files to export
        :type file_path: str
        :param variables_list: list of the constraints (func name)
        :type variables_list: list(str)
        :param extension: file extension
        :type extension: str
        """
        fig = plt.figure(figsize=(11, 6))
        plt.xlabel("Iterations", fontsize=12)
        dspace = self.opt_problem.design_space
        desvars = dspace.variables_names

        has_dv = False
        for data_name in data_list:
            if data_name in desvars:
                has_dv = True
                break
        dv_hist = None
        if has_dv:
            dv_hist = array(self.database.get_x_history())
        for data_name in data_list:
            if data_name in desvars:
                mask = []
                for currvar in dspace.variables_names:
                    size = dspace.variables_sizes[currvar]
                    isvar = data_name == currvar
                    mask += [isvar] * size

                data_hist = dv_hist[:, mask]
            else:
                data_hist = self.database.get_func_history(data_name, x_hist=False)
            labels = data_name

            data_hist = array(data_hist).real
            plt.plot(arange(len(data_hist)), data_hist, label=labels)

        plt.title("History plot")
        plt.legend()
        self._save_and_show(
            fig, save=save, show=show, file_path=file_path, extension=extension
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Distribution factory
====================

This module contains a factory to instantiate a :class:`.Distribution`
from its class name.
The class can be internal to |g|
or located in an external module whose path is provided to the constructor.
It also provides a list of available distributions types
and allows you to test if a distribution type is available.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

from gemseo.core.factory import Factory
from gemseo.uncertainty.distributions.distribution import Distribution

standard_library.install_aliases()


from gemseo import LOGGER


class DistributionFactory(object):
    """This factory instantiates a :class:`.Distribution` from its class name.
    The class can be internal to |g| or located in an external module
    whose path is provided to the constructor.
    """

    def __init__(self):
        """
        Initializes the factory: scans the directories to search for
        subclasses of Distribution.
        Searches in "GEMSEO_PATH" and gemseo.uncertainty.p_dist.
        """
        self.factory = Factory(Distribution, ("gemseo.uncertainty.distributions",))

    def create(self, distribution_name, variable, **parameters):
        """
        Creates a distribution.

        :param str distribution_name: name of the distribution (its classname)
        :param str variable: variable name.
        :param parameters: distribution parameters
        :return: distribution_name distribution
        """
        return self.factory.create(distribution_name, variable=variable, **parameters)

    @property
    def distributions(self):
        """
        Lists the available classes.

        :returns: the list of classes names.
        :rtype: list(str)
        """
        return self.factory.classes

    def is_available(self, distribution_name):
        """
        Checks the availability of a distribution.

        :param str distribution_name:  name of the distribution.
        :returns: True if the distribution is available.
        :rtype: bool
        """
        return self.factory.is_available(distribution_name)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Distributions based on scipy
============================

Overview
--------

The :class:`.SPDistribution` class is a concrete class
inheriting from :class:`.Distribution` which is an abstract one.
SP stands for `scipy <https://docs.scipy.org/doc/scipy/reference/
tutorial/stats.html>`_ which is the library it relies on.

Construction
------------

The :class:`.SPDistribution` of a given uncertain variable is built
from mandatory arguments:

- a variable name,
- a distribution name recognized by OpenTURNS,
- a set of parameters provided as a dictionary
  of keyword arguments named as the arguments of the scipy constructor
  of this distribution.

.. warning::

    The distribution parameters must be provided according to the signature
    of the scipy classes. `Access the scipy documentation
    <https://docs.scipy.org/doc/scipy/reference/stats.html>`_.

The constructor has also optional arguments:

- a variable dimension (default: 1),
- a standard representation of these parameters
  (default: use :code:`parameters`).

Classical distributions
-----------------------

This module also implements a deliberately limited selection
of classical probability distributions
in a user-friendly way: :class:`.SPExponentialDistribution`,
:class:`.SPNormalDistribution`, :class:`.SPTriangularDistribution`,
and :class:`.SPUniformDistribution`. More precisely,
the argument whose nature is a dictionary of keyword parameters
is replaced with several user-defined keyword arguments.
In this way, the use writes :code:`SPUniformDistribution('x', -1., 3.)`
or :code:`SPUniformDistribution('x', lower=-1., upper=3.)`
instead of :code:`SPDistribution('x', 'Normal', (-1., 3.))`.
Furthermore, these classes inheriting from :class:`.SPDistribution`
are documented in such a way that a newbie could easily apprehend them.
"""
from __future__ import absolute_import, division, unicode_literals

import scipy.stats as sp_stats
from future import standard_library
from numpy import array, vstack

from gemseo.uncertainty.distributions.distribution import Distribution
from gemseo.uncertainty.distributions.sp_cdist import SPComposedDistribution

standard_library.install_aliases()

from gemseo import LOGGER

SP_WEBSITE = "https://docs.scipy.org/doc/scipy/reference/stats.html"


class SPDistribution(Distribution):

    """The SPDistribution inherits from Distribution. It relies on the
    probabilistic modeling features of the scipy library.
    """

    COMPOSED_DISTRIBUTION = SPComposedDistribution

    def __init__(
        self, variable, distribution, parameters, dimension=1, standard_parameters=None
    ):
        """Constructor

        :param str variable: variable name.
        :param str distribution: distribution name.
        :param tuple parameters: distribution parameters.
        :param int dimension: variable dimension.
        :param dict standard_parameters: standard parameters.
        """
        super(SPDistribution, self).__init__(
            variable, distribution, parameters, dimension, standard_parameters
        )
        self.marginals = self.__create_distributions(
            self.distribution_name, self.parameters
        )
        math_support = str(self.support)
        LOGGER.info("|_ Mathematical support: %s ", math_support)
        num_range = str(self.range)
        LOGGER.info("|_ Numerical range: %s", num_range)

    def get_sample(self, n_samples=1):
        """Get several samples.

        :param int n_samples: number of samples.
        :return: samples
        :rtype: array
        """
        return vstack([marginal.rvs(n_samples) for marginal in self.marginals]).T

    def cdf(self, vector):
        """Evaluate the cumulative density function of the random variable
        marginals for a given instance.

        :param array vector: instance of the random variable.
        :return: cdf values
        :rtype: array
        """
        return array(
            [self.marginals[index].cdf(value) for index, value in enumerate(vector)]
        )

    def inverse_cdf(self, vector):
        """Evaluate the inverse of the cumulative density function of the
        random variable marginals for a given unit vector .

        :param array vector: vector of values comprised between 0 and 1 with
            same dimension as the random variable.
        :return: inverse cdf values
        :rtype: array
        """
        return array(
            [self.marginals[index].ppf(value) for index, value in enumerate(vector)]
        )

    @property
    def mean(self):
        """Get the mean of the random variable.

        :return: mean of the random variable.
        :rtype: array
        """
        mean = [marginal.mean() for marginal in self.marginals]
        return array(mean)

    @property
    def standard_deviation(self):
        """Get the standard deviation of the random variable.

        :return: standard deviation of the random variable.
        :rtype: array
        """
        std = [marginal.std() for marginal in self.marginals]
        return array(std)

    def __create_distributions(self, distribution, parameters):
        """For each variable dimension, instantiate an openturns distribution
        from its class name and parameters.

        :param str distribution: distribution name,
            ie its openturns class name.
        :param parameters: parameters of the scipy distribution.
        :return: scipy distribution.
        """
        try:
            sp_dist = getattr(sp_stats, distribution)
        except Exception:
            raise ValueError("%s is an unknown scipy " "distribution. " % distribution)
        try:
            sp_dist(**parameters)
        except Exception:
            raise ValueError(
                "%s does not take these arguments. "
                "More details on: %s" % (distribution, SP_WEBSITE)
            )
        sp_dist = [sp_dist(**parameters)] * self.dimension
        self.__set_bounds(sp_dist)
        return sp_dist

    def __set_bounds(self, distributions, extrema_level=1e-12):
        """Set mathematical and numerical bounds (= support and range).

        :param distributions: list of scipy distributions.
        :param float extremal_level: quantile level corresponding to the lower
            bound of the numerical random variable range. Default: 1e-12.
        """
        self.math_lower_bound = []
        self.math_upper_bound = []
        self.num_lower_bound = []
        self.num_upper_bound = []
        for distribution in distributions:
            dist_range = distribution.interval(1.0)
            self.math_lower_bound.append(dist_range[0])
            self.math_upper_bound.append(dist_range[1])
            l_b = distribution.ppf(extrema_level)
            u_b = distribution.ppf(1 - extrema_level)
            self.num_lower_bound.append(l_b)
            self.num_upper_bound.append(u_b)
        self.math_lower_bound = array(self.math_lower_bound)
        self.math_upper_bound = array(self.math_upper_bound)
        self.num_lower_bound = array(self.num_lower_bound)
        self.num_upper_bound = array(self.num_upper_bound)

    def _pdf(self, index):
        """Get the probability density function of a marginal.

        :param int index: marginal index.
        :return: probability density function
        :rtype: function
        """

        def pdf(point):
            """Probability Density Function (PDF).

            :param float point: point value.
            :return: PDF value.
            :rtype: float
            """
            return self.marginals[index].pdf(point)

        return pdf

    def _cdf(self, index):
        """Get the cumulative density function of a marginal.

        :param int index: marginal index.
        :return: cumulative density function
        :rtype: function
        """

        def cdf(level):
            """Cumulative Density Function (CDF).

            :param float point: point value.
            :return: CDF value.
            :rtype: float
            """
            return self.marginals[index].cdf(level)

        return cdf


class SPNormalDistribution(SPDistribution):
    """ Create a normal distribution. """

    def __init__(self, variable, mu=0.0, sigma=1.0, dimension=1):
        """Constructor.

        :param str variable: variable name.
        :param float mu: mean.
        :param float sigma: standard deviation.
        :param int dimension: dimension.
        """
        standard_parameters = {self.MU: mu, self.SIGMA: sigma}
        parameters = {"loc": mu, "scale": sigma}
        super(SPNormalDistribution, self).__init__(
            variable, "norm", parameters, dimension, standard_parameters
        )


class SPUniformDistribution(SPDistribution):
    """ Create a uniform distribution. """

    def __init__(self, variable, lower=0.0, upper=1.0, dimension=1):
        """Constructor.

        :param str variable: variable name.
        :param float lower: lower bound.
        :param float upper: upper bound.
        :param int dimension: dimension.
        """
        parameters = {"loc": lower, "scale": upper - lower}
        standard_parameters = {self.LOWER: lower, self.UPPER: upper}
        super(SPUniformDistribution, self).__init__(
            variable, "uniform", parameters, dimension, standard_parameters
        )


class SPTriangularDistribution(SPDistribution):
    """ Create a triangular distribution. """

    def __init__(self, variable, lower=0.0, mode=0.5, upper=1.0, dimension=1):
        """Constructor.

        :param str variable: variable name.
        :param float lower: lower bound.
        :param float mode: mode.
        :param float upper: upper bound.
        :param int dimension: dimension.
        """
        parameters = {
            "loc": lower,
            "scale": upper - lower,
            "c": (mode - lower) / float(upper - lower),
        }
        standard_parameters = {self.LOWER: lower, self.MODE: mode, self.UPPER: upper}
        super(SPTriangularDistribution, self).__init__(
            variable, "triang", parameters, dimension, standard_parameters
        )


class SPExponentialDistribution(SPDistribution):
    """ Create a exponential distribution. """

    def __init__(self, variable, rate=1.0, loc=0.0, dimension=1):
        """Constructor.

        :param str variable: variable name.
        :param float rate: rate parameter.
        :param float loc: location parameter.
        :param int dimension: dimension.
        """
        parameters = {"loc": loc, "scale": 1 / float(rate)}
        super(SPExponentialDistribution, self).__init__(
            variable, "expon", parameters, dimension
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Base class for all optimization post processings
************************************************
"""
from __future__ import absolute_import, division, unicode_literals

import inspect
from os.path import abspath, dirname, exists, join, splitext

import matplotlib
import pylab
from future import standard_library

from gemseo.core.grammar import InvalidDataException
from gemseo.core.json_grammar import JSONGrammar
from gemseo.utils.source_parsing import SourceParsing

standard_library.install_aliases()


class OptPostProcessor(object):
    """Abstract class for optimization post processing methods"""

    def __init__(self, opt_problem):
        """
        Constructor

        :param opt_problem: the optimization problem to run
        """
        self.opt_problem = opt_problem
        self.database = opt_problem.database
        cls_name = self.__class__.__name__
        name = cls_name + "_options"
        f_class = inspect.getfile(self.__class__)
        comp_dir = abspath(dirname(f_class))
        schema_file = join(comp_dir, name + ".json")
        if not exists(schema_file):
            schema_file = join(comp_dir, "options", name + ".json")
        if not exists(schema_file):
            raise ValueError(
                "Options grammar for optimization "
                + " post processor does not exists"
                + ", expected: "
                + str(schema_file)
                + "or: "
                + str(join(comp_dir, name + ".json"))
            )
        self.opt_grammar = JSONGrammar(name, schema_file=schema_file)
        if hasattr(self.__class__, "_plot"):
            descr_opts = SourceParsing.get_options_doc(
                self.__class__._plot
            )  # pylint: disable=E1101
            self.opt_grammar.add_description(descr_opts)
        if hasattr(self.__class__, "_run"):
            descr_opts = SourceParsing.get_options_doc(self.__class__._run)
            self.opt_grammar.add_description(descr_opts)
        # the data dict to eventually rebuild the plot in another framework
        self.out_data_dict = {}
        # Store output files generated by the post processing
        self.output_files = []

    def execute(self, **options):
        """
        Executes the post processing

        :param options: options dict, see associated JSON file
        """
        self.check_options(**options)
        if not self.opt_problem.database:
            raise ValueError(
                "Optimization problem was not solved,"
                + " cannot run post processing "
                + str(self.__class__.__name__)
            )
        self._run(**options)

    def check_options(self, **options):
        """
        Checks the options of the post processor

        :param options: options dict, see associated JSON file
        """
        try:
            self.opt_grammar.load_data(options)
        except InvalidDataException:
            name = self.__class__.__name__
            raise InvalidDataException(
                "Invalid options for post processing " + name + " got : " + str(options)
            )

    def _run(self, **options):
        """Visualizes the optimization history

        :param options: options of the plot,
            see associated JSON file
        """
        self._plot(**options)

    def _plot(self, **options):
        """Visualizers

        :param options: options dict, see associated JSON file

        """
        raise NotImplementedError()

    def _save_and_show(self, fig, file_path, save=True, show=False, extension="pdf"):
        """
        Saves figures and or shows it depending on options

        :param fig: matplotlib figure
        :param file_path: file path to save the plot. If it does not end with
            the specified extension, the extension is replaced
        :param save: save the figure (default: True)
        :param show: show the figure (default: False)
        :param extension: file extension (default: 'pdf')
        """
        matplotlib.rcParams.update({"font.size": 10})
        if save:
            if not file_path.endswith(extension):
                basen = splitext(file_path)[0]
                file_path = basen + "." + extension
            fig.savefig(file_path, bbox_inches="tight")
            self.output_files.append(file_path)
        if show:
            try:
                pylab.plt.show(fig)
            except TypeError:
                pylab.plt.show()



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
YvsX plot
==========

A :class:`.YvsX` plot represents samples of a couple :math:`(x,y)`
as a set of points whose values are stored in a :class:`.Dataset`. The user
can select the style of line or markers, as well as the color.
"""
from __future__ import absolute_import, division, unicode_literals

import matplotlib.pyplot as plt
from future import standard_library

from gemseo.post.dataset.dataset_plot import DatasetPlot

standard_library.install_aliases()


class YvsX(DatasetPlot):
    """ Plot curve y versus x. """

    def _plot(self, x, y, x_comp=0, y_comp=0, style="o", color="blue"):
        """Surface.

        :param x: name of the variable on the x-axis
        :type x: str
        :param y: name of the variable on the y-axis
        :type z: str
        :param x_comp: x component. Default: 0.
        :type x_comp: int
        :param y_comp: y component. Default: 0.
        :type y_comp: int
        :param style: line style. Default: 'o'.
        :type style: str
        :param color: point color. Default: 'blue'.
        :type color: str
        """
        x_data = self.dataset[x][x][:, x_comp]
        y_data = self.dataset[y][y][:, y_comp]

        fig = plt.figure()
        axes = fig.add_subplot(1, 1, 1)
        axes.plot(x_data, y_data, style, color=color)
        if self.dataset.sizes[x] == 1:
            axes.set_xlabel(x)
        else:
            axes.set_xlabel(x + "(" + str(x_comp) + ")")
        if self.dataset.sizes[y] == 1:
            axes.set_ylabel(y)
        else:
            axes.set_ylabel(y + "(" + str(y_comp) + ")")
        fig = plt.gcf()
        return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Fitting a distribution from data based on OpenTURNS
===================================================

Overview
--------

The :class:`.OTDistributionFitter` class considers several samples
of an uncertain variable, fits a user-defined probability distribution
from this dataset and returns a :class:`.OTDistribution`.
It can also return a goodness-of-fit measure
associated with this distribution,
e.g. Bayesian Information Criterion, Kolmogorov test or Chi Squared test,
or select an optimal distribution among a collection according to
a criterion with a threshold.

Construction
------------

The :class:`.OTDistributionFitter` of a given uncertain variable is built
from only two arguments:

- a variable name,
- a one-dimensional numpy array.

Capabilities
------------

Fit a distribution
~~~~~~~~~~~~~~~~~~

The :meth:`.OTDistributionFitter.fit` method takes a distribution name
recognized by OpenTURNS as argument (e.g. 'Normal', 'Uniform', 'Exponential',
...) as argument and returns an :class:`.OTDistribution`
whose underlying OpenTURNS distribution is the specified one fitted
from the dataset passed to the constructor.

Measure the goodness-of-fit
~~~~~~~~~~~~~~~~~~~~~~~~~~~

The :meth:`.OTDistributionFitter.measure` method has two mandatory arguments:

- a distribution which is a either a :class:`.OTDistribution`
  or a distribution name from which :meth:`!fit` method
  builds a :class:`.OTDistribution`,
- a fitting criterion name.

.. note::

   Use the :meth:`.OTDistributionFitter.get_available_criteria` method to get
   the complete list of available criteria
   and the :meth:`.OTDistributionFitter.get_significance_tests` method
   to get the list of available criteria which are significance tests.

The :meth:`.OTDistributionFitter.measure` method can also use a level
associated with the criterion.

The :meth:`.OTDistributionFitter.measure` methods returns a goodness-of-fit
measure whose nature is either a scalar
when the criterion is not a significance test
or a tuple when the criterion is a significance test. In that case,
the first component of the tuple is a boolean indicating if the measured
distribution is acceptable to model the data and the second one is
a dictionary containing the test statistics, the p-value and
the significance level.

Select an optimal distribution
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The :meth:`.OTDistributionFitter.select` method select aims to select an
optimal distribution among a collection. It uses two mandatory arguments:

- a list of distribution, either a list of distributions names
  or a list of :class:`.OTDistribution`,
- a fitting criterion name.

The :meth:`.OTDistributionFitter.select` method can also use a level
associated with the criterion and a criterion selection:

- 'best': select the distribution minimizing (or maximizing, depending
  on the criterion) the criterion,
- 'first': Select the first distribution for which the criterion is
  greater (or lower, depending on the criterion) than the level.
"""
from __future__ import absolute_import, division, unicode_literals

import openturns as ots
from future import standard_library
from numpy import ndarray
from past.builtins import basestring

from gemseo.uncertainty.distributions.ot_dist import OTDistribution

standard_library.install_aliases()

from gemseo import LOGGER


class OTDistributionFitter(object):
    """ OpenTURNS distribution fitter. """

    AVAILABLE_FACTORIES = {}
    for factory in ots.DistributionFactory.GetContinuousUniVariateFactories():
        factory_class_name = factory.getImplementation().getClassName()
        dist_name = factory_class_name.split("Factory")[0]
        AVAILABLE_FACTORIES[dist_name] = getattr(ots, factory_class_name)

    AVAILABLE_FITTING_TESTS = {
        "BIC": ots.FittingTest.BIC,
        "Kolmogorov": ots.FittingTest.Kolmogorov,
        "ChiSquared": ots.FittingTest.ChiSquared,
    }

    SIGNIFICANCE_TESTS = ["Kolmogorov", "ChiSquared"]

    CRITERIA_TO_MAXIMIZE = []
    CRITERIA_TO_MINIMIZE = ["BIC"]

    def __init__(self, variable, data):
        """Constructor.

        :param str variable: variable name.
        :param array data: data.
        """
        self.variable = variable
        try:
            isinstance(data, ndarray)
            self.data = ots.Sample(data.reshape((-1, 1)))
        except AttributeError:
            raise TypeError("data must be a numpy array")

    def _get_factory(self, distribution):
        """Get distribution factory.

        :param str distribution: distribution name.
        """
        try:
            distribution_factory = self.AVAILABLE_FACTORIES[distribution]
        except KeyError:
            distributions = ", ".join(list(self.AVAILABLE_FACTORIES.keys()))
            raise ValueError(
                "{} is not a name of "
                "distribution available for fitting. "
                "Available ones are: {}.".format(distribution, distributions)
            )
        return distribution_factory

    def _get_fitting_test(self, criterion):
        """Get fitting test.

        :param str criterion: criterion name.
        """
        try:
            fitting_test = self.AVAILABLE_FITTING_TESTS[criterion]
        except KeyError:
            tests = ", ".join(list(self.AVAILABLE_FITTING_TESTS.keys()))
            raise ValueError(
                criterion + " is not a name of fitting test. "
                "Available ones are: " + tests + "."
            )
        return fitting_test

    def fit(self, distribution):
        """Fit a distribution.

        :param str distribution: distribution name.
        """
        factory = self._get_factory(distribution)
        fitted_distribution = factory().build(self.data)
        parameters = fitted_distribution.getParameter()
        distribution = OTDistribution(self.variable, distribution, parameters)
        return distribution

    def measure(self, distribution, criterion, level=0.05):
        """Measure the goodness-of-fit of a distribution to data.

        :param distribution: distribution.
        :type distribution: OTDistribution or str
        :param str fitting_criterion: goodness-of-fit criterion.
        :param float level: risk of committing a Type 1 error,
            that is an incorrect rejection of a true null hypothesis,
            for criteria based on test hypothesis.
            Default: 0.05.
        """
        if isinstance(distribution, basestring):
            distribution = self.fit(distribution)
        if distribution.dimension > 1:
            raise TypeError("A 1D distribution is required.")
        distribution = distribution.marginals[0]
        fitting_test = self._get_fitting_test(criterion)
        if criterion in self.SIGNIFICANCE_TESTS:
            result = fitting_test(self.data, distribution, level)
            details = {
                "p-value": result.getPValue(),
                "statistics": result.getStatistic(),
                "level": level,
            }
            result = (result.getBinaryQualityMeasure(), details)
        else:
            result = fitting_test(self.data, distribution)
        return result

    def select(
        self, distributions, fitting_criterion, level=0.05, selection_criterion="best"
    ):
        """Select the best distribution.

        :param distributions: list of distributions.
        :type distribution: list(OTDistribution) or list(str)
        :param str fitting_criterion: goodness-of-fit criterion.
        :param float level: significance level. For hypothesis tests,
            this is the risk of committing a Type 1 error,
            that is an incorrect rejection of a true null hypothesis.
            For other tests, this is a threshold.
            Default: 0.05.
        :param str selection_criterion: selection criterion
        """
        results = []
        for index, distribution in enumerate(distributions):
            if isinstance(distribution, basestring):
                distribution = self.fit(distribution)
            results.append(self.measure(distribution, fitting_criterion, level))
            distributions[index] = distribution
        index = self.select_from_results(
            results, fitting_criterion, level, selection_criterion
        )
        return distributions[index]

    @classmethod
    def select_from_results(
        cls, results, fitting_criterion, level=0.05, selection_criterion="best"
    ):
        """Select the best distribution from results

        :param list results: results
        :param str fitting_criterion: goodness-of-fit criterion.
        :param float level: significance level. For hypothesis tests,
            this is the risk of committing a Type 1 error,
            that is an incorrect rejection of a true null hypothesis.
            For other tests, this is a threshold.
            Default: 0.05.
        :param str selection_criterion: selection criterion
        """
        if fitting_criterion in cls.SIGNIFICANCE_TESTS:
            for index, _ in enumerate(results):
                results[index] = results[index][1]["p-value"]
            if sum([pval > level for pval in results]) == 0:
                LOGGER.warning(
                    "All criteria values are lower than the " "significance level %s.",
                    level,
                )
        if selection_criterion == "best" or level is None:
            index = cls.__find_opt_distribution(results, fitting_criterion)
        else:
            index = cls.__apply_first_strategy(results, fitting_criterion, level)
        return index

    @classmethod
    def __apply_first_strategy(cls, results, fitting_criterion, level=0.05):
        """Select the index of the best distribution from results
        by applying the "first" strategy.

        :param dict results: results
        :param str fitting_criterion: goodness-of-fit criterion.
        :param float level: significance level. For hypothesis tests,
            this is the risk of committing a Type 1 error,
            that is an incorrect rejection of a true null hypothesis.
            For other tests, this is a threshold.
            Default: 0.05.
        """
        select = False
        index = 0
        for result in results:
            select = result >= level
            if select:
                break
            index += 1
        if not select:
            index = cls.__find_opt_distribution(results, fitting_criterion)
        return index

    @classmethod
    def __find_opt_distribution(cls, results, fitting_criterion):
        """ Returns the index of the optimum distribution."""
        if fitting_criterion in cls.CRITERIA_TO_MINIMIZE:
            index = results.index(min(results))
        else:
            index = results.index(max(results))
        return index

    def get_available_distributions(self):
        """ Get available distributions. """
        return sorted(self.AVAILABLE_FACTORIES.keys())

    def get_available_criteria(self):
        """ Get available goodness-of-fit criteria. """
        return sorted(self.AVAILABLE_FITTING_TESTS.keys())

    def get_significance_tests(self):
        """ Get significance tests. """
        return sorted(self.SIGNIFICANCE_TESTS)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""
Basic display of optimization history : functions, and x
********************************************************
"""
from __future__ import absolute_import, division, unicode_literals

import matplotlib.gridspec as gridspec
import pylab
from future import standard_library
from matplotlib.colors import SymLogNorm
from matplotlib.ticker import LogFormatter, MaxNLocator
from numpy import abs as np_abs
from numpy import append, arange, argmin, array, atleast_2d, concatenate, hstack, isnan
from numpy import log10 as np_log10
from numpy import logspace
from numpy import max as np_max
from numpy import min as np_min
from numpy import ones, ones_like
from numpy import sort as np_sort
from numpy import vstack, where
from numpy.linalg import norm

from gemseo.core.function import MDOFunction
from gemseo.post.core.colormaps import PARULA, RG_SEISMIC
from gemseo.post.core.hessians import SR1Approx
from gemseo.post.opt_post_processor import OptPostProcessor

standard_library.install_aliases()
from gemseo import LOGGER


class OptHistoryView(OptPostProcessor):
    """
    The **OptHistoryView** post processing
    performs separated plots:
    the design variables history,
    the objective function history,
    the history of hessian approximation of the objective,
    the inequality constraint history,
    the equality constraint history,
    and constraints histories.

    By default, all design variables are considered.
    A sublist of design variables can be passed as options.
    Minimum and maximum values for the plot can be passed as options.
    The objective function can also be represented in terms of difference
    w.r.t. the initial value
    It is possible either to save the plot, to show the plot or both.
    """

    def __init__(self, opt_problem):
        """
        Constructor

        :param opt_problem: the optimization problem to run
        """
        super(OptHistoryView, self).__init__(opt_problem)
        self.cmap = PARULA  # "viridis"  # "jet"
        self.ineq_cstr_cmap = RG_SEISMIC  # "seismic" "PRGn_r"
        self.eq_cstr_cmap = "seismic"  # "seismic" "PRGn_r"

    def _plot(
        self,
        show=False,
        save=False,
        file_path=None,
        variables_names=None,
        obj_min=None,
        obj_max=None,
        obj_relative=False,
        extension="pdf",
    ):
        """
        Plots the optimization history:
        1 plot for the design variables
        1 plot for the objective function
        1 plot for the Hessian approximation of the objective
        1 plot for inequality constraints
        1 plot for equality constraints

        :param show: if True, displays the plot windows
        :type show: bool
        :param save: if True, exports plot to pdf
        :type save: bool
        :param file_path: the base paths of the files to export
        :type file_path: str
        :param variables_names: list of the names of the variables to display
        :type variables_names: list(str)
        :param obj_max: maximum value for the objective in the plot
        :type obj_max: float
        :param obj_min: minimum value for the objective in the plot
        :type obj_min: float
        :param obj_relative: plot the objective value difference
            with the initial value
        :type obj_relative: bool
        :param extension: file extension
        :type extension: str
        """
        # compute the names of the inequality and equality constraints
        ineq_cstr = self.opt_problem.get_ineq_constraints()
        ineq_cstr_names = [c.name for c in ineq_cstr]
        eq_cstr = self.opt_problem.get_eq_constraints()
        eq_cstr_names = [c.name for c in eq_cstr]

        obj_name = self.opt_problem.get_objective_name()
        obj_history, x_history, n_iter = self._get_history(obj_name, variables_names)

        # design variables
        self._create_variables_plot(
            x_history, variables_names, save, show, file_path, extension
        )

        # objective function
        self._create_obj_plot(
            obj_history,
            n_iter,
            save,
            show,
            file_path,
            obj_min,
            obj_max,
            obj_relative=obj_relative,
            extension=extension,
        )

        self._create_x_star_plot(
            x_history, n_iter, save, show, file_path=file_path, extension=extension
        )

        # Hessian plot
        plot_hessian = not self.database.is_func_grad_history_empty(obj_name)
        if plot_hessian:
            self._create_hessian_approx_plot(
                self.database, obj_name, save, show, file_path, extension
            )

        # inequality and equality constraints
        self._plot_cstr_history(
            ineq_cstr_names, MDOFunction.TYPE_INEQ, save, show, file_path, extension
        )
        self._plot_cstr_history(
            eq_cstr_names, MDOFunction.TYPE_EQ, save, show, file_path, extension
        )

    def _plot_cstr_history(
        self, cstr_names, cstr_type, save, show, file_path, extension
    ):
        """
        Create the plot for (in)equality constraints

        :param cstr_names: names of the constraints
        :param cstr_type: type of the constraints in {MDOFunction.TYPE_INEQ,
            MDOFunction.TYPE_EQ}
        :param show: if True, displays the plot windows
        :param save: if True, exports plot to pdf
        :param file_path: the base paths of the files to export
        :param extension: file extension
        """
        if cstr_names is not None:
            _, constraints_history = self._get_constraints(cstr_names)
            self._create_cstr_plot(
                constraints_history,
                cstr_type,
                cstr_names,
                save,
                show,
                file_path,
                extension,
            )

    def _get_history(self, fname, variables_names):
        """
        Access the optimization history of a function and the design
        variables at which it was computed

        :param fname: name of the function
        :param variables_names: list of the names of the variables to display
        :returns: list of function values
            list of design variables values
            list of iterations
        """
        f_hist, x_hist = self.database.get_func_history(fname, x_hist=True)
        f_hist = array(f_hist).real
        x_hist = array(x_hist).real

        if variables_names is not None:
            # select only the interesting columns
            block_list = []
            column = 0
            for var in self.opt_problem.design_space.variables_names:
                if var in variables_names:
                    size = self.opt_problem.design_space.variables_sizes[var]
                    block_list.append(x_hist[:, column : column + size])
                    column += size
            # concatenate the blocks
            x_hist = hstack(block_list)
        n_iter = x_hist.shape[0]
        return f_hist, x_hist, n_iter

    def _get_constraints(self, cstr_names):
        """
        Extracts a history of constraints

        :param cstr_names: names of the constraints
        :returns: bounds of the constraints and history array
        """
        available_data_names = self.database.get_all_data_names()
        for cstr_name in cstr_names:
            if cstr_name not in available_data_names:
                cstr_names.remove(cstr_name)

        constraints_history = []
        bnd_list = -1e300 * ones(len(cstr_names))

        for cstr_i, cstr_name in enumerate(cstr_names):

            cstr_history = array(self.database.get_func_history(cstr_name)).real

            constraints_history.append(cstr_history)
            bnd_list[cstr_i] = max(
                bnd_list[cstr_i],
                max(abs(np_min(cstr_history)), abs(np_max(cstr_history))),
            )

        return bnd_list, constraints_history

    def _normalize_x_hist(self, x_history, variables_names):
        """
        Normalizes the design variables history

        :param x_history: the history for the design variables
        :param variables_names: list of the names of the variables to display
        :returns: the normalized design variables array
        """
        x_hist_n = x_history.copy()
        lower_bounds = self.opt_problem.design_space.get_lower_bounds(variables_names)
        upper_bounds = self.opt_problem.design_space.get_upper_bounds(variables_names)
        norm_coeff = 1 / (np_abs(upper_bounds - lower_bounds))
        for i in range(x_history.shape[0]):
            x_hist_n[i, :] = (x_hist_n[i, :] - lower_bounds) * norm_coeff
        return x_hist_n

    def _create_variables_plot(
        self, x_history, variables_names, save, show, file_path=None, extension="pdf"
    ):
        """
        Creates the design variables plot

        :param x_history: the design variables history
        :param variables_names: list of the names of the variables to display
        :param n_variables: number of parameters
        :param save: saves the plot to a file
        :param show: shows the matplotlib figure
        :param file_path: the base paths of the files to export
        :param extension: file extension
        """
        if len(x_history) < 2:
            return
        n_variables = x_history.shape[1]
        norm_x_history = self._normalize_x_hist(x_history, variables_names)

        fig = pylab.plt.figure(figsize=(11, 6))
        grid = gridspec.GridSpec(1, 2, width_ratios=[15, 1], wspace=0.04, hspace=0.6)

        # design variables
        ax1 = fig.add_subplot(grid[0, 0])
        im1 = ax1.imshow(
            norm_x_history.T,
            cmap=self.cmap,
            interpolation="nearest",
            vmin=0.0,
            vmax=1.0,
            aspect="auto",
        )
        # y ticks: names of the variables
        design_space = self.opt_problem.design_space
        y_labels = []
        if variables_names is None:
            variables_names = design_space.variables_names
        for variable_name in variables_names:
            size = design_space.variables_sizes[variable_name]
            name = variable_name
            if size > 1:
                name += " (0)"
            y_labels.append(name)
            for i in range(1, size):
                y_labels.append("(" + str(i) + ")")

        ax1.set_yticklabels(y_labels)
        ax1.set_yticks(arange(n_variables))
        # ax1.invert_yaxis()

        ax1.set_title("Evolution of the optimization variables")
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))

        # colorbar
        ax2 = fig.add_subplot(grid[0, 1])
        fig.colorbar(im1, cax=ax2)

        if save:
            if file_path is not None:
                path = file_path + "_variables_history." + extension
            else:
                path = "design_variables_history." + extension
            pylab.plt.savefig(path)
            self.output_files.append(path)

        # Set window size
        mng = pylab.plt.get_current_fig_manager()
        mng.resize(700, 1000)

        if show:
            pylab.plt.show()
        else:
            pylab.plt.close(fig)

    def _create_obj_plot(
        self,
        obj_history,
        n_iter,
        save,
        show,
        file_path=None,
        obj_min=None,
        obj_max=None,
        obj_relative=False,
        extension="pdf",
    ):
        """
        Creates the design variables plot

        :param obj_history: history of the objective function
        :param n_iter: number of iterations
        :param save: saves the plot to the disc
        :param show: shows the matplotlib figures
        :param file_path: the base paths of the files to export
        :param obj_max: maximum value for the objective in the plot
        :param obj_min: minimum value for the objective in the plot
        :param obj_relative: plot the objective value difference
            with the initial value
        :param extension: file extension
        """
        # if max problem, plot -objective value
        if not self.opt_problem.minimize_objective:
            obj_history = -obj_history
        if obj_relative:
            LOGGER.info(
                "Plot of optimization history "
                "with relative variation compared to "
                "initial point objective value = %s",
                obj_history[0],
            )
            obj_history -= obj_history[0]
        # Remove nans
        n_x = len(obj_history)
        x_absc = arange(n_x)
        x_absc_nan = None
        idx_nan = isnan(obj_history)

        if idx_nan.size > 0:
            obj_history = obj_history[~idx_nan]
            x_absc_nan = x_absc[idx_nan]
            x_absc = x_absc[~idx_nan]

        fmin = np_min(obj_history)
        fmax = np_max(obj_history)

        fig = pylab.plt.figure(figsize=(11, 6))
        # objective function
        pylab.plt.xlabel("Iterations", fontsize=12)
        pylab.plt.ylabel("Objective value", fontsize=12)

        pylab.plt.plot(x_absc, obj_history)

        if idx_nan.size > 0:
            for x_i in x_absc_nan:
                pylab.plt.axvline(x_i, color="purple")

        if obj_min is not None and obj_min < fmin:
            fmin = obj_min
        if obj_max is not None and obj_max > fmax:
            fmax = obj_max
        pylab.plt.ylim([fmin, fmax])
        pylab.plt.xlim([0, n_iter])
        ax1 = fig.gca()
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))
        pylab.plt.grid(True)
        pylab.plt.title("Evolution of the objective value")

        if save:
            if file_path is not None:
                path = file_path + "_obj_history." + extension
            else:
                path = "objective_function_history." + extension
            pylab.plt.savefig(path)
            self.output_files.append(path)

        # Set window size
        mng = pylab.plt.get_current_fig_manager()
        mng.resize(700, 1000)

        if show:
            pylab.plt.show()
        else:
            pylab.plt.close(fig)

    def _create_x_star_plot(
        self, x_history, n_iter, save, show, file_path=None, extension="pdf"
    ):
        """
        Creates the design variables plot

        :param x_history: history of the design variables
        :param x_opt: x optimum
        :param n_iter: number of iterations
        :param save: save the figure
        :param show: show the figure
        :param file_path: the base paths of the files to export
        :param extension: file extension
        """
        fig = pylab.plt.figure(figsize=(11, 6))
        # objective function
        pylab.plt.xlabel("Iterations", fontsize=12)
        pylab.plt.ylabel("||x-x*||", fontsize=12)

        n_i = x_history.shape[0]
        _, x_opt, _, _, _ = self.opt_problem.get_optimum()

        normalize = self.opt_problem.design_space.normalize_vect

        x_xstar = [
            norm(normalize(x_history[i, :]) - normalize(x_opt)) for i in range(n_i)
        ]
        # Draw a vertical line at the optimum
        ind_opt = argmin(x_xstar)
        pylab.plt.axvline(x=ind_opt, color="r")
        pylab.plt.semilogy(arange(len(x_xstar)), x_xstar)
        # ======================================================================
        # try:
        #     pylab.plt.semilogy(np.arange(len(x_xstar)), x_xstar)
        # except ValueError:
        #     LOGGER.warning("Cannot use log scale for x_star plot since" +
        #                    "all values are not positive !")
        # ======================================================================
        ax1 = fig.gca()
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))
        pylab.plt.grid(True)
        pylab.plt.title("Distance to the optimum")
        pylab.plt.xlim([0, n_iter])

        if save:
            if file_path is not None:
                path = file_path + "_x_xstar_history." + extension
            else:
                path = "x_xstar_history." + extension
            pylab.plt.savefig(path)
            self.output_files.append(path)

        # Set window size
        mng = pylab.plt.get_current_fig_manager()
        mng.resize(700, 1000)

        if show:
            pylab.plt.show()
        else:
            pylab.plt.close(fig)

    @staticmethod
    def _cstr_number(cstr_names, cstr_history):
        """
        Computes the total scalar constraints number

        :param cstr_names: names of the constraints
        :param cstr_history: history of the constraints
        :returns: number of constraints
        """
        n_cstr = 0
        for cstr_i in range(len(cstr_names)):
            c_hist_loc = atleast_2d(cstr_history[cstr_i]).T
            if c_hist_loc.shape[1] == 1:
                c_hist_loc = c_hist_loc.T
            n_cstr += c_hist_loc.shape[0]
        LOGGER.debug("Total constraints number =%s", n_cstr)
        return n_cstr

    def _create_cstr_plot(
        self,
        cstr_history,
        cstr_type,
        cstr_names,
        save,
        show,
        file_path=None,
        extension="pdf",
    ):
        """
        Creates the constraints plot: 1 line per constraint component

        :param cstr_history: history of the constraints
        :param cstr_names: names of the constraints
        :param save: if True, exports plot to pdf
        :returns: the plot
        :param file_path: the base paths of the files to export
        :param extension: file extension
        """
        n_cstr = self._cstr_number(cstr_names, cstr_history)
        if n_cstr == 0:
            return

        # matrix of all constraints' values
        cstr_matrix = None
        vmax = 0.0
        cstr_labels = []

        max_iter = 0
        for (i, cstr_history_i) in enumerate(cstr_history):
            history_i = atleast_2d(cstr_history_i).T
            if history_i.shape[1] == 1:
                history_i = history_i.T

            if max_iter < history_i.shape[1]:
                max_iter = history_i.shape[1]

        for (i, cstr_history_i) in enumerate(cstr_history):
            history_i = atleast_2d(cstr_history_i).T
            if history_i.shape[1] == 1:
                history_i = history_i.T

            nb_components = history_i.shape[0]

            if history_i.shape[1] == max_iter:  # TEST
                for component_j in range(nb_components):
                    # compute the label of the constraint
                    if component_j == 0:
                        cstr_name = cstr_names[i]
                        if nb_components >= 2:
                            cstr_name += " (" + str(component_j + 1) + ")"
                        cstr_labels.append(cstr_name)
                    else:
                        cstr_labels.append("(" + str(component_j + 1) + ")")

                    history_i_j = atleast_2d(history_i[component_j, :])

                    # max value
                    notnans = ~isnan(history_i_j)
                    vmax = max(vmax, np_max(np_abs(history_i_j[notnans])))

                    # build the constraint matrix
                    if cstr_matrix is None:
                        cstr_matrix = history_i_j
                    else:
                        cstr_matrix = vstack((cstr_matrix, history_i_j))

        fig = self._build_cstr_fig(cstr_matrix, cstr_type, vmax, n_cstr, cstr_labels)

        if save:
            if file_path is not None:
                path = file_path + "_" + cstr_type + "_constraints_history." + extension
            else:
                path = cstr_type + "_constraints_history." + extension
            pylab.plt.savefig(path)
            self.output_files.append(path)
        if show:
            pylab.plt.show()
        else:
            pylab.plt.close(fig)

    def _build_cstr_fig(self, cstr_matrix, cstr_type, vmax, n_cstr, cstr_labels):
        """
        Builds the constraints figure

        :param cstr_matrix : matrix of constraints values
        :param cstr_labels: labels for the constraints
        :param vmax: maximum constraint absolute value
        :param n_cstr: number of constraints
        :param cstr_labels : labels of constraints names
        :returns: the matplotlib figure
        """
        # cmap of the constraints
        fullname = "equality"
        if cstr_type == MDOFunction.TYPE_EQ:
            cmap = self.eq_cstr_cmap
        else:
            cmap = self.ineq_cstr_cmap
            fullname = "in" + fullname

        idx_nan = isnan(cstr_matrix)
        hasnan = idx_nan.any()
        if hasnan > 0:
            cstr_matrix[idx_nan] = 0.0

        # generation of the image
        fig = pylab.plt.figure(figsize=(11, 6))
        grid = gridspec.GridSpec(1, 2, width_ratios=[15, 1], wspace=0.04, hspace=0.6)
        ax1 = fig.add_subplot(grid[0, 0])
        im1 = ax1.imshow(
            cstr_matrix,
            cmap=cmap,
            interpolation="nearest",
            vmin=-vmax,
            vmax=vmax,
            aspect="auto",
            norm=SymLogNorm(linthresh=1.0, vmin=-vmax, vmax=vmax),
        )
        if hasnan > 0:
            x_absc_nan = where(idx_nan.any(axis=0))[0]
            for x_i in x_absc_nan:
                pylab.plt.axvline(x_i, color="purple")

        ax1.tick_params(labelsize=9)
        ax1.set_yticks(list(range(n_cstr)))
        ax1.set_yticklabels(cstr_labels)

        ax1.set_xlabel("evaluations")
        ax1.set_title("Evolution of the " + fullname + " constraints")
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))

        ax1.hlines(
            list(range(len(cstr_matrix))),
            [-0.5],
            [len(cstr_matrix[0]) - 0.5],
            alpha=0.1,
            lw=0.5,
        )

        # color map
        cax = fig.add_subplot(grid[0, 1])
        if 0.0 < vmax < 1.0:
            thick_min = int(np_log10(vmax))
        else:
            thick_min = 0
        if vmax > 1.0:
            thick_max = int(np_log10(vmax))
        else:
            thick_max = 0
        thick_num = thick_max - thick_min + 1
        levels_pos = logspace(thick_min, thick_max, num=thick_num, base=10.0)
        if vmax != 0.0:
            levels_pos = np_sort(append(levels_pos, vmax))
        levels_neg = np_sort(-levels_pos)
        levels_neg = append(levels_neg, 0)
        levels = concatenate((levels_neg, levels_pos))
        col_bar = fig.colorbar(im1, cax=cax, ticks=levels)
        col_bar.ax.tick_params(labelsize=9)
        cax.set_xlabel("symlog")

        mng = pylab.plt.get_current_fig_manager()
        # mng.full_screen_toggle()
        mng.resize(700, 1000)
        return fig

    def _create_hessian_approx_plot(
        self, history, obj_name, save, show, file_path=None, extension="pdf"
    ):
        """
        Creates the plot of the Hessian approximation

        :param history: the optimization history
        :param obj_name: the function name
        :param extension: file extension
        """
        try:
            approximator = SR1Approx(history)
            _, diag, _, _ = approximator.build_approximation(
                funcname=obj_name, save_diag=True
            )
            diag = [ones_like(diag[0])] + diag  # Add first iteration blank
            diag = array(diag).T
        except ValueError as err:
            LOGGER.warning("Failed to create Hessian approximation: %s", str(err))
            return

        # if max problem, plot -Hessian
        if not self.opt_problem.minimize_objective:
            diag = -diag

        fig = pylab.plt.figure(figsize=(11, 6))
        grid = gridspec.GridSpec(1, 2, width_ratios=[15, 1], wspace=0.04, hspace=0.6)
        # matrix
        axe = fig.add_subplot(grid[0, 0])

        axe.set_title("Hessian diagonal approximation")
        axe.set_xlabel("Iterations", fontsize=12)
        axe.set_ylabel("Variable id", fontsize=12)
        axe.xaxis.set_major_locator(MaxNLocator(integer=True))
        vmax = max(abs(np_max(diag)), abs(np_min(diag)))
        linthresh = 10 ** ((np_log10(vmax) - 5.0))
        img = axe.imshow(
            diag.real,
            cmap=self.cmap,
            interpolation="nearest",
            aspect="auto",
            norm=SymLogNorm(linthresh=linthresh, vmin=-vmax, vmax=vmax),
        )
        axe.invert_yaxis()

        # colorbar
        vmax = max(abs(np_max(diag)), abs(np_min(diag)))
        thick_min = int(np_log10(linthresh))
        thick_max = int(np_log10(vmax))
        thick_num = thick_max - thick_min + 1
        levels_pos = logspace(thick_min, thick_max, num=thick_num, base=10.0)
        levels_pos = append(levels_pos, vmax)
        levels_neg = np_sort(-levels_pos)
        levels_neg = append(levels_neg, 0)
        levels = concatenate((levels_neg, levels_pos))

        l_f = LogFormatter(base=10, labelOnlyBase=False)
        cax = fig.add_subplot(grid[0, 1])
        cax.invert_yaxis()
        fig.colorbar(img, cax=cax, ticks=levels, format=l_f)

        mng = pylab.plt.get_current_fig_manager()
        # mng.full_screen_toggle()
        mng.resize(700, 1000)
        if save:
            path = "hessian_approx." + extension
            if file_path is not None:
                path = file_path + "_" + path
            pylab.plt.savefig(path)
            self.output_files.append(path)
        if show:
            pylab.plt.show()
        else:
            pylab.plt.close(fig)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
R2 error measure
================

The :mod:`~gemseo.mlearning.qual_measure.r2_measure` module
implements the concept of R2 measures for machine learning algorithms.

This concept is implemented through the :class:`.R2Measure` class
and overloads the :meth:`!MLErrorMeasure._compute_measure` method.

The R2 is defined by

.. math::

    R_2(\\hat{y}) = 1 - \\frac{\\sum_i (\\hat{y}_i - y_i)^2}
                              {\\sum_i (y_i-\\bar{y})^2},

where
:math:`\\hat{y}` are the predictions,
:math:`y` are the data points and
:math:`\\bar{y}` is the mean of :math:`y`.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from sklearn.metrics import r2_score

from gemseo.mlearning.qual_measure.error_measure import MLErrorMeasure

standard_library.install_aliases()


class R2Measure(MLErrorMeasure):
    """ R2 measure for machine learning. """

    def _compute_measure(self, outputs, predictions, multioutput=True):
        """Compute MSE.

        :param ndarray outputs: reference outputs.
        :param ndarray predictions: predicted outputs.
        :param bool multioutput: if True, return the error measure for each
            output component. Otherwise, average these errors. Default: True.
        :return: MSE value.
        """
        multioutput = "raw_values" if multioutput else "uniform_average"
        return r2_score(outputs, predictions, multioutput=multioutput)



# -*- coding: utf-8 -*-
# Based on unittest2/plugins/junitxml.py,
# which is itself based on the junitxml plugin from py.test distributed
# under the MIT license:
# The MIT License (MIT)
#
# Copyright (c) 2004-2016 Holger Krekel and others
#
# Permission is hereby granted, free of charge,
# to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons
# to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice
# shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Output test reports in junit-xml format.
# This plugin implements:func:`startTest`,:func:`testOutcome` and
# :func:`stopTestRun` to compile and then output a test report in
# junit-xml format. By default, the report is written to a file called
# ``nose2-junit.xml`` in the current working directory.
# You can configure the output filename by setting ``path``
# in a ``[junit-xml]``
# section in a config file.  Unicode characters which are invalid in XML 1.0
# are replaced with the ``U+FFFD`` replacement character.
# In the case that your
# software throws an error with an invalid byte string.
# By default, the ranges of discouraged characters are replaced as well.
#  This can be changed by setting the ``keep_restricted``
#   configuration variable to ``True``.

"""
Output test reports in junit-xml format
***************************************
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import functools

from future import standard_library

standard_library.install_aliases()


from gemseo import LOGGER
__unittest = True  # pylint: disable=C0103


def link_to(*dec_args):
    """Prints the requirements covered by the test.

    :param dec_args: decorator arguments
    """
    def decorator(function):
        """Decorates the function to print requirement

        :param function: returns: the decorated function
        :returns: the decorated function
        """
        @functools.wraps(function)
        def call_and_log_req(*args, **kwargs):
            """Logs the covered requirement
            then calls the function

            :param args: function args
            :param kwargs: function kwargs
            :param args:
            :param kwargs:
            :returns: function output
            """
            LOGGER.info('[Covers requirements: %s]', ', '.join(dec_args))
            return function(*args, **kwargs)
        # Add attribute to returned function
        call_and_log_req.requirements = dec_args
        return call_and_log_req
    return decorator



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Principal component dimension reduction algorithm
=================================================

The :class:`PCA` class wraps the PCA from Scikit-learn.

Dependence
----------
This dimension reduction algorithm relies on the PCA class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.decomposition.PCA.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import sqrt
from sklearn.decomposition import PCA as SKLPCA

from gemseo.mlearning.transform.dimension_reduction.dimension_reduction import (
    DimensionReduction,
)

standard_library.install_aliases()


class PCA(DimensionReduction):
    """ Principal component dimension reduction algorithm. """

    def __init__(self, name="PCA", n_components=5, **parameters):
        """Constructor.

        :param str name: transformer name. Default: 'PCA'.
        :param int n_components: number of components. Default: 5.
        :param parameters: Optional parameters for sklearn PCA constructor.
        """
        super(PCA, self).__init__(name, n_components=n_components, **parameters)
        self.algo = SKLPCA(n_components, **parameters)

    def fit(self, data):
        """Fit transformer to data.

        :param ndarray data: data to be fitted.
        """
        self.algo.fit(data)

    def transform(self, data):
        """Transform data.

        :param ndarray data: data  to be transformed.
        :return: transformed data.
        :rtype: ndarray
        """
        return self.algo.transform(data)

    def inverse_transform(self, data):
        """Perform an inverse transform on the data.

        :param ndarray data: data to be inverse transformed.
        :return: inverse transformed data.
        :rtype: ndarray
        """
        return self.algo.inverse_transform(data)

    def compute_jacobian(self, data):
        """Compute Jacobian of the pca transform.

        :param ndarray data: data where the Jacobian is to be computed.
        :return: Jacobian matrix.
        :rtype: ndarray
        """
        return self.algo.components_

    def compute_jacobian_inverse(self, data):
        """Compute Jacobian of the pca inverse_transform.

        :param ndarray data: data where the Jacobian is to be computed.
        :return: Jacobian matrix.
        :rtype: ndarray
        """
        return self.algo.components_.T

    @property
    def components(self):
        """ Components """
        return sqrt(self.algo.singular_values_) * self.algo.components_.T



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Abstract dataset plot
=====================

The :mod:`~gemseo.post.dataset.dataset_plot` module implements the abstract
:class:`.DatasetPlot` class
whose purpose is to build a graphical representation
of a :class:`.Dataset` and to display it on screen or save it to a file.
This abstract class has to be overloaded by concrete ones implementing
at least method :meth:`!DatasetPlot._run`.
"""
from __future__ import absolute_import, division, unicode_literals

from numbers import Number

import matplotlib
import pylab
from future import standard_library
from six import string_types

standard_library.install_aliases()


class DatasetPlot(object):
    """ Abstract plot class for dataset. """

    def __init__(self, dataset):
        """Constructor.

        :param Dataset dataset: dataset
        """
        if dataset.is_empty():
            raise ValueError("Dataset is empty.")
        self.dataset = dataset
        self.output_files = []

    def execute(
        self, save=True, show=False, file_path=None, extension="pdf", **plot_options
    ):
        """
        Executes the post processing

        :param show: if True, displays the plot windows. Default: False.
        :type show: bool
        :param save: if True, exports plot to pdf. Default: False.
        :type save: bool
        :param file_path: the base paths of the files to export. Default: None.
        :type file_path: str
        :param extension: file extension. Default: 'pdf'.
        :type extension: str
        :param plot_options: options passed to the _plot() method.
        """
        self._run(save, show, file_path, extension, **plot_options)

    def _run(self, save, show, file_path, extension, **plot_options):
        """Define run of the post processing, calling _plot() and saving
        or displaying figure.

        :param plot_options: options passed to the _plot() method.
        :param show: if True, displays the plot windows. Default: False.
        :type show: bool
        :param save: if True, exports plot to pdf. Default: False.
        :type save: bool
        :param file_path: the base paths of the files to export. Default: None.
        :type file_path: str
        :param extension: file extension. Default: 'pdf'.
        :type extension: str
        :param plot_options: options passed to the _plot() method.
        """
        fig = self._plot(**plot_options)
        if isinstance(fig, list):
            for index, subfig in enumerate(fig):
                subfig.tight_layout()
                filename = self.__class__.__name__ + str(index)
                self._save_and_show(
                    subfig,
                    filename=filename,
                    save=save,
                    show=show,
                    file_path=file_path,
                    extension=extension,
                )
        else:
            fig.tight_layout()
            self._save_and_show(
                fig,
                filename=self.__class__.__name__,
                save=save,
                show=show,
                file_path=file_path,
                extension=extension,
            )

    def _plot(self, **options):
        """Define plot of the post processing,
        to be implemented in subclasses

        :param options: plotting options according to associated json file
        """
        raise NotImplementedError

    def _save_and_show(
        self, fig, filename, save=True, show=False, file_path=None, extension="pdf"
    ):
        """
        Saves figures and or shows it depending on options

        :param fig: matplotlib figure
        :param str filename: name of the file to save the plot
        :param bool save: save the figure (default: True)
        :param bool show: show the figure (default: False)
        :param str file_path: file path to save the plot (default: None)
        :param str extension: file extension (default: 'pdf')
        """
        matplotlib.rcParams.update({"font.size": 10})
        if save:
            file_path = file_path or self.dataset.name
            fpath = file_path + "_" + filename
            fpath += "." + extension
            fig.savefig(fpath, bbox_inches="tight")
            if fpath not in self.output_files:
                self.output_files.append(fpath)
        if show:
            try:
                pylab.plt.show(fig)
            except TypeError:
                pylab.plt.show()

    def _get_varnames(self, df_columns):
        """Get varnames from columns of a pandas dataframe.

        :param list(tuple) dataframe: pandas dataframe
        """
        new_columns = []
        for column in df_columns:
            if self.dataset.sizes[column[1]] == 1:
                new_columns.append(column[1])
            else:
                new_columns.append(column[1] + "(" + str(column[2]) + ")")
        return new_columns

    def _get_label(self, varname):
        """Returns string label associated to a variable name,
        as well as refactored variable name.

        :param varname: variable name, either a string
            or a (str, int) tuple.
        """
        if isinstance(varname, string_types):
            label = varname
            varname = (self.dataset.get_group(varname), varname, "0")
        elif hasattr(varname, "__len__") and len(varname) == 3:
            is_string = isinstance(varname[0], string_types)
            is_string = is_string and isinstance(varname[1], string_types)
            is_number = isinstance(varname[2], Number)
            if is_string and is_number:
                label = varname[1] + "(" + str(varname[2]) + ")"
                varname[2] = str(varname[2])
                varname = tuple(varname)
            else:
                raise TypeError(
                    "varname must be either a string or a tuple"
                    " whose first component is a string and second"
                    " one is an integer."
                )
        else:
            raise TypeError(
                "varname must be either a string or a tuple"
                " whose first component is a string and second"
                " one is an integer."
            )
        return label, varname
